import os
from opyrant import (configure, states, panels, behavior, stimuli, blocks, subjects)


def run(configuration_file):
    """ Run an experiment or set of experiments detailed in the configuration file """

    # Check for configuration file
    if not os.path.exists(configuration_file):
        raise IOError("Configuration file could not be found: %s" % configuration_file)

    # Load the configuration based on its extension
    extension = os.path.splitext(configuration_file)[1].lower()
    if extension == ".yaml":
        parameters = configure.ConfigureYAML(configuration_file)
    elif extension == ".json":
        parameters = configure.ConfigureJSON(configuration_file)
    else:
        raise ValueError("Configuration file must be either yaml or json")

    # Set up subject

    # Set up panel

    # Set up experiments

    # Set up stimlulus conditions

    # Set up blocks

    # Set up BlockHandler



from numpy import random

class BaseSchedule(object):
    """Maintains logic for deciding whether to consequate trials.

    This base class provides the most basic reinforcent schedule: every
    response is consequated.

    Methods:
    consequate(trial) -- returns a boolean value based on whether the trial
        should be consequated. Always returns True.

    """

    def __init__(self):
        super(BaseSchedule, self).__init__()

    def consequate(self,trial):
        assert hasattr(trial, 'correct') and isinstance(trial.correct, bool)
        if trial.correct:
            return True
        else:
            return True

class ContinuousReinforcement(BaseSchedule):
    """Maintains logic for deciding whether to consequate trials.

    This base class provides the most basic reinforcent schedule: every
    response is consequated.

    Methods:
    consequate(trial) -- returns a boolean value based on whether the trial
        should be consequated. Always returns True.

    """
    def __init__(self):
        super(ContinuousReinforcement, self).__init__()

    def consequate(self,trial):
        assert hasattr(trial, 'correct') and isinstance(trial.correct, bool)
        if trial.correct:
            return True
        else:
            return True

class FixedRatioSchedule(BaseSchedule):
    """Maintains logic for deciding whether to consequate trials.

    This class implements a fixed ratio schedule, where a reward reinforcement
    is provided after every nth correct response, where 'n' is the 'ratio'.

    Incorrect trials are always reinforced.

    Methods:
    consequate(trial) -- returns a boolean value based on whether the trial
        should be consequated.

    """
    def __init__(self, ratio=1):
        super(FixedRatioSchedule, self).__init__()
        self.ratio = max(ratio,1)
        self._update()

    def _update(self):
        self.cumulative_correct = 0
        self.threshold = self.ratio

    def consequate(self,trial):
        assert hasattr(trial, 'correct') and isinstance(trial.correct, bool)
        if trial.correct==True:
            self.cumulative_correct += 1
            if self.cumulative_correct >= self.threshold:
                self._update()
                return True
            else:
                return False
        elif trial.correct==False:
            self.cumulative_correct = 0
            return True
        else:
            return False

    def __unicode__(self):
        return "FR%i" % self.ratio

class VariableRatioSchedule(FixedRatioSchedule):
    """Maintains logic for deciding whether to consequate trials.

    This class implements a variable ratio schedule, where a reward
    reinforcement is provided after every a number of consecutive correct
    responses. On average, the number of consecutive responses necessary is the
    'ratio'. After a reinforcement is provided, the number of consecutive
    correct trials needed for the next reinforcement is selected by sampling
    randomly from the interval [1,2*ratio-1]. e.g. a ratio of '3' will require
    consecutive correct trials of 1, 2, 3, 4, & 5, randomly.

    Incorrect trials are always reinforced.

    Methods:
    consequate(trial) -- returns a boolean value based on whether the trial
        should be consequated.


    """
    def __init__(self, ratio=1):
        super(VariableRatioSchedule, self).__init__(ratio=ratio)

    def _update(self):
        ''' update min correct by randomly sampling from interval [1:2*ratio)'''
        self.cumulative_correct = 0
        self.threshold = random.randint(1, 2*self.ratio)

    def __unicode__(self):
        return "VR%i" % self.ratio

class PercentReinforcement(BaseSchedule):
    """Maintains logic for deciding whether to consequate trials.

    This class implements a probabalistic reinforcement, where a reward reinforcement
    is provided x percent of the time.

    Incorrect trials are always reinforced.

    Methods:
    consequate(trial) -- returns a boolean value based on whether the trial
        should be consequated.

    """
    def __init__(self, prob=1):
        super(PercentReinforcement, self).__init__()
        self.prob = prob

    def consequate(self,trial):
        assert hasattr(trial, 'correct') and isinstance(trial.correct, bool)
        if trial.correct:
            return random.random() < self.prob
        else:
            return True

    def __unicode__(self):
        return "PR%i" % self.prob

SCHEDULE_DICT = dict(continuous=ContinuousReinforcement,
                     fixed=FixedRatioSchedule,
                     fixedratio=FixedRatioSchedule,
                     variable=VariableRatioSchedule,
                     variableratio=VariableRatioSchedule,
                     percent=PercentReinforcement)



import numpy as np
from scipy.stats import norm
from scipy.stats import beta
from matplotlib import mlab

# d-prime
def dprime(confusion_matrix):
    """
    Function takes in a 2x2 confusion matrix and returns the d-prime value for the predictions.

    d' = z(hit rate)-z(false alarm rate)

    http://en.wikipedia.org/wiki/D'
    """
    if max(confusion_matrix.shape) > 2:
        return False
    else:
        hit_rate = confusion_matrix[0, 0] / confusion_matrix[0, :].sum()
        fa_rate = confusion_matrix[1, 0] / confusion_matrix[1, :].sum()

        nudge = 0.0001
        if (hit_rate >= 1): hit_rate = 1 - nudge
        if (hit_rate <= 0): hit_rate = 0 + nudge
        if (fa_rate >= 1): fa_rate = 1 - nudge
        if (fa_rate <= 0): fa_rate = 0 + nudge

        dp = norm.ppf(hit_rate)-norm.ppf(fa_rate)
        return dp


# accuracy (% correct)
def acc(confusion_matrix):
    """Function takes in a NxN confusion matrix
    and returns the fraction of correct predictions"""

    x = confusion_matrix.diagonal().sum()
    N = confusion_matrix.sum()
    p = x / N

    return p


# accuracy (% correct)
def acc_ci(confusion_matrix, alpha=0.05):
    """Function takes in a NxN confusion matrix
    and returns the fraction of correct predictions"""

    x = confusion_matrix.diagonal().sum()
    N = confusion_matrix.sum()

    ci = beta.interval(1-alpha,x,N-x)
    return ci


# matthew's correlation coefficient
def mcc(confusion_matrix):
    """Function takes in a 2x2 confusion matrix
    and returns the Matthew's Correlation Coefficient for the predictions.

    MCC = (TP*TN-FP*FN)/sqrt((TP+FP)*(TP+FN)*(TN+FP)*(TN+FN))

    http://en.wikipedia.org/wiki/Matthews_correlation_coefficient

    """
    if max(confusion_matrix.shape) > 2:
        return False
    else:
        true_pos = confusion_matrix[0, 0]
        true_neg = confusion_matrix[1, 1]
        false_pos = confusion_matrix[1, 0]
        false_neg = confusion_matrix[0, 1]
        a = (true_pos*true_neg-false_pos*false_neg)/np.sqrt((true_pos+false_pos)*(true_pos+false_neg)*(true_neg+false_pos)*(true_neg+false_neg))
        return a

def create_conf_matrix(expected, predicted):
    """
    Function takes in a 1-D array of expected values and a 1-D array of predictions
    and returns a confusion matrix with size corresponding to the number of classes.

    http://en.wikipedia.org/wiki/Confusion_matrix

    Keyword arguments:
    expected  -- list of expected or true values
    predicted -- list of predicted or response values

    Returns the confusion matrix as a numpy array m[expectation,prediction]

    """
    n_classes = max(len(set(expected)), len(set(predicted)), 2)

    m = np.zeros((n_classes, n_classes))
    for exp, pred in zip(expected, predicted):
        m[exp,pred] += 1
    return m

class Performance():
    """ use this to compute performance metrics """
    def __init__(self, expected, predicted):
        self.confusion_matrix = create_conf_matrix(expected, predicted)
    def n_classes(self):
        return max(self.confusion_matrix.shape)
    def dprime(self):
        return dprime(self.confusion_matrix)
    def acc(self):
        return acc(self.confusion_matrix)
    def acc_ci(self):
        return acc_ci(self.confusion_matrix)
    def mcc(self):
        return mcc(self.confusion_matrix)


class Session(object):
    """docstring for Session"""
    def __init__(self, arg):
        super(Session, self).__init__()
        self.arg = arg

        
        


import wave
import sys
import struct
import time
import subprocess
import threading
import traceback
import shlex
import os
import fnmatch
import string
import random
import datetime as dt
import numpy as np
import scipy as sp
import scipy.special
from contextlib import closing
from argparse import ArgumentParser
from opyrant import Error


try:
    import simplejson as json
except ImportError:
    import json

class NumpyAwareJSONEncoder(json.JSONEncoder):
    """ this json encoder converts numpy arrays to lists so that json can write them.

    example usage:

    >>> import numpy as np
    >>> dict_to_save = {'array': np.zeros((5,))}
    >>> json.dumps(dict_to_save,
                   cls=NumpyAwareJSONEncoder
                   )
    '{"array": [0.0, 0.0, 0.0, 0.0, 0.0]}'

    """

    def default(self, obj):
        if isinstance(obj, np.ndarray):
                return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def filter_files(directory, file_pattern="*", recursive=False):
    """ Finds all files in directory that match a specific pattern.

    Parameters
    ----------
    directory: string
        File path to the directory to search
    file_pattern: string
        A glob to filter on using fnmatch.filter
    recursive: bool
        Whether or not to search subdirectories

    Returns
    -------
    list of matching files
    """

    if not os.path.isdir(directory):
        raise IOError("%s is not a directory" % directory)

    files = list()
    for rootdir, dirname, fnames in os.walk(directory):
        matches = fnmatch.filter(fnames, file_pattern)
        files.extend(os.path.join(rootdir, fname) for fname in matches)
        if not recursive:
            dirname[:] = list()

    return files


# consider importing this from python-neo
class Event(object):
    """docstring for Event"""
    def __init__(self, time=None, duration=None, label='', name=None, description=None, file_origin=None, *args, **kwargs):
        super(Event, self).__init__()
        self.time = time
        self.duration = duration
        self.label = label
        self.name = name
        self.description = description
        self.file_origin = file_origin
        self.annotations = {}
        self.annotate(**kwargs)

    def annotate(self,**kwargs):
        self.annotations.update(kwargs)


class Stimulus(Event):
    """docstring for Stimulus"""
    def __init__(self, *args, **kwargs):
        super(Stimulus, self).__init__(*args, **kwargs)
        if self.label=='':
            self.label = 'stimulus'

class AuditoryStimulus(Stimulus):
    """docstring for AuditoryStimulus"""
    def __init__(self, *args, **kwargs):
        super(AuditoryStimulus, self).__init__(*args, **kwargs)
        if self.label=='':
            self.label = 'auditory_stimulus'


def run_state_machine(start_in='pre', error_state=None, error_callback=None, **state_functions):
    """runs a state machine defined by the keyword arguments

    >>> def run_start():
    >>>    print "in 'run_start'"
    >>>    return 'next'
    >>> def run_next():
    >>>    print "in 'run_next'"
    >>>    return None
    >>> run_state_machine(start_in='start',
    >>>                   start=run_start,
    >>>                   next=run_next)
    in 'run_start'
    in 'run_next'
    None
    """
    # make sure the start state has a function to run
    assert (start_in in state_functions.keys())
    # make sure all of the arguments passed in are callable
    for func in state_functions.values():
        assert hasattr(func, '__call__')

    state = start_in
    while state is not None:
        try:
            state = state_functions[state]()
        except Exception, e:
            if error_callback:
                error_callback(e)
                raise
            else:
                raise
            state = error_state


class Trial(Event):
    """docstring for Trial"""
    def __init__(self,
                 index=None,
                 type_='normal',
                 class_=None,
                 *args, **kwargs):
        super(Trial, self).__init__(*args, **kwargs)
        self.label = 'trial'
        self.session = None
        self.index = index
        self.type_ = type_
        self.stimulus = None
        self.class_ = class_
        self.response = None
        self.correct = None
        self.rt = None
        self.reward = False
        self.punish = False
        self.events = []
        self.stim_event = None


class Command(object):
    """
    Enables to run subprocess commands in a different thread with TIMEOUT option.

    via https://gist.github.com/kirpit/1306188

    Based on jcollado's solution:
    http://stackoverflow.com/questions/1191374/subprocess-with-timeout/4825933#4825933

    """
    command = None
    process = None
    status = None
    output, error = '', ''

    def __init__(self, command):
        if isinstance(command, basestring):
            command = shlex.split(command)
        self.command = command

    def run(self, timeout=None, **kwargs):
        """ Run a command then return: (status, output, error). """
        def target(**kwargs):
            try:
                self.process = subprocess.Popen(self.command, **kwargs)
                self.output, self.error = self.process.communicate()
                self.status = self.process.returncode
            except:
                self.error = traceback.format_exc()
                self.status = -1
        # default stdout and stderr
        if 'stdout' not in kwargs:
            kwargs['stdout'] = subprocess.PIPE
        if 'stderr' not in kwargs:
            kwargs['stderr'] = subprocess.PIPE
        # thread
        thread = threading.Thread(target=target, kwargs=kwargs)
        thread.start()
        thread.join(timeout)
        if thread.is_alive():
            self.process.terminate()
            thread.join()
        return self.status, self.output, self.error

def parse_commandline(arg_str=sys.argv[1:]):
    """ parse command line arguments
    note: optparse is depreciated w/ v2.7 in favor of argparse

    """
    parser=ArgumentParser()
    parser.add_argument('-B', '--box',
                      action='store', type=int, dest='box', required=False,
                      help='(int) box identifier')
    parser.add_argument('-S', '--subject',
                      action='store', type=str, dest='subj', required=False,
                      help='subject ID and folder name')
    parser.add_argument('-c','--config',
                      action='store', type=str, dest='config_file', default='config.json', required=True,
                      help='configuration file [default: %(default)s]')
    args = parser.parse_args(arg_str)
    return vars(args)

def check_cmdline_params(parameters, cmd_line):
    # if someone is using red bands they should ammend the checks I perform here
    allchars=string.maketrans('','')
    nodigs=allchars.translate(allchars, string.digits)
    if not ('box' not in cmd_line or cmd_line['box'] == int(parameters['panel_name'].encode('ascii','ignore').translate(allchars, nodigs))):
        print "box number doesn't match config and command line"
        return False
    if not ('subj' not in cmd_line or int(cmd_line['subj'].encode('ascii','ignore').translate(allchars, nodigs)) == int(parameters['subject'].encode('ascii','ignore').translate(allchars, nodigs))):
        print "subject number doesn't match config and command line"
        return False
    return True



def time_in_range(start, end, x):
    """Return true if x is in the range [start, end]"""
    if start <= end:
        return start <= x <= end
    else:
        return start <= x or x <= end

def is_day((latitude, longitude) = ('32.82', '-117.14')):
    """Is it daytime?

    (lat,long) -- latitude and longitude of location to check (default is San Diego*)
    Returns True if it is daytime

    * Discovered by the Germans in 1904, they named it San Diego,
    which of course in German means a whale's vagina. (Burgundy, 2004)
    """
    import ephem
    obs = ephem.Observer()
    obs.lat = latitude # San Diego, CA
    obs.long = longitude
    sun = ephem.Sun()
    sun.compute()
    next_sunrise = ephem.localtime(obs.next_rising(sun))
    next_sunset = ephem.localtime(obs.next_setting(sun))
    return next_sunset < next_sunrise


def check_time(schedule,fmt="%H:%M"):
    """ determine whether trials should be done given the current time and the light schedule

    returns Boolean if current time meets schedule

    schedule='sun' will change lights according to local sunrise and sunset

    schedule=[('07:00','17:00')] will have lights on between 7am and 5pm
    schedule=[('06:00','12:00'),('18:00','24:00')] will have lights on between

    """
    if schedule == 'sun':
        if is_day():
            return True
    elif schedule == "night":
        if not is_day():
            return True
    else:
        for epoch in schedule:
            assert len(epoch) is 2
            now = dt.datetime.time(dt.datetime.now())
            start = dt.datetime.time(dt.datetime.strptime(epoch[0],fmt))
            end = dt.datetime.time(dt.datetime.strptime(epoch[1],fmt))
            if time_in_range(start,end,now):
                return True
    return False

def wait(secs=1.0, final_countdown=0.0,waitfunc=None):
    """Smartly wait for a given time period.

    secs -- total time to wait in seconds
    final_countdown -- time at end of secs to wait and constantly poll the clock
    waitfunc -- optional function to run in a loop during hogCPUperiod

    If secs=1.0 and final_countdown=0.2 then for 0.8s python's time.sleep function will be used,
    which is not especially precise, but allows the cpu to perform housekeeping. In
    the final hogCPUsecs the more precise method of constantly polling the clock
    is used for greater precision.
    """
    #initial relaxed period, using sleep (better for system resources etc)
    if secs > final_countdown:
        time.sleep(secs-final_countdown)
        secs = final_countdown # only this much is now left

    #It's the Final Countdown!!
    #hog the cpu, checking time
    t0 = time.time()
    while (time.time()-t0) < secs:
        #let's see if any events were collected in meantime
        try:
            waitfunc()
        except:
            pass

def auditory_stim_from_wav(wav):
    with closing(wave.open(wav,'rb')) as wf:
        (nchannels, sampwidth, framerate, nframes, comptype, compname) = wf.getparams()

        duration = float(nframes)/sampwidth
        duration = duration * 2.0 / framerate
        stim = AuditoryStimulus(time=0.0,
                                duration=duration,
                                name=wav,
                                label='wav',
                                description='',
                                file_origin=wav,
                                annotations={'nchannels': nchannels,
                                             'sampwidth': sampwidth,
                                             'framerate': framerate,
                                             'nframes': nframes,
                                             'comptype': comptype,
                                             'compname': compname,
                                             }
                                )
    return stim

def concat_wav(input_file_list, output_filename='concat.wav'):
    """ concat a set of wav files into a single wav file and return the output filename

    takes in a tuple list of files and duration of pause after the file

    input_file_list = [
        ('a.wav', 0.1),
        ('b.wav', 0.09),
        ('c.wav', 0.0),
        ]

    returns a list of AuditoryStimulus objects

    TODO: add checks for sampling rate, number of channels, etc.
    """

    cursor = 0
    epochs = [] # list of file epochs
    audio_data = ''
    with closing(wave.open(output_filename, 'wb')) as output:
        for input_filename, isi in input_file_list:

            # read in the wav file
            with closing(wave.open(input_filename,'rb')) as wav_part:
                try:
                    params = wav_part.getparams()
                    output.setparams(params)
                    fs = output.getframerate()
                except: # TODO: what was I trying to except here? be more specific
                    pass

                audio_frames = wav_part.readframes(wav_part.getnframes())

            # append the audio data
            audio_data += audio_frames

            part_start = cursor
            part_dur = len(audio_frames)/params[1]

            epochs.append(AuditoryStimulus(time=float(part_start)/fs,
                                           duration=float(part_dur)/fs,
                                           name=input_filename,
                                           file_origin=input_filename,
                                           annotations=params,
                                           label='motif'
                                           ))
            cursor += part_dur # move cursor length of the duration

            # add isi
            if isi > 0.0:
                isi_frames = ''.join([struct.pack('h', fr) for fr in [0]*int(fs*isi)])
                audio_data += isi_frames
                cursor += len(isi_frames)/params[1]

        # concat all of the audio together and write to file
        output.writeframes(audio_data)


    description = 'concatenated on-the-fly'
    concat_wav = AuditoryStimulus(time=0.0,
                                  duration=epochs[-1].time+epochs[-1].duration,
                                  name=output_filename,
                                  label='wav',
                                  description=description,
                                  file_origin=output_filename,
                                  annotations=output.getparams(),
                                  )

    return (concat_wav,epochs)


def get_num_open_fds():
    '''
    return the number of open file descriptors for current process

    .. warning: will only work on UNIX-like os-es.
    '''

    pid = os.getpid()
    procs = subprocess.check_output(
        [ "lsof", '-w', '-Ff', "-p", str( pid ) ] )

    nprocs = len(
        filter(
            lambda s: s and s[ 0 ] == 'f' and s[1: ].isdigit(),
            procs.split( '\n' ) )
        )
    return nprocs

def rand_from_log_shape_dist(alpha=10):
    """
    randomly samples from a distribution between 0 and 1 with pdf shaped like the log function
    low probability of getting close to zero, increasing probability going towards 1
    alpha determines how sharp the curve is, higher alpha, sharper curve.
    """
    beta = (alpha + 1) * np.log(alpha + 1) - alpha
    t = random.random()
    ret = ((beta * t-1)/(sp.special.lambertw((beta*t-1)/np.e)) - 1) / alpha
    return max(min(np.real(ret), 1), 0)



# useful exception for going into sleep
class GoodNite(Exception):
    """ exception for when the lights should be off """
    pass

class EndExperiment(Exception):
    """ exception for when an experiment should terminate"""
    pass

class EndSession(Exception):
    """ exception for when a session should terminate """
    pass

class EndBlock(Exception):
    """ exception for when a block should terminate """
    pass

## defining Error classes for operant HW control
class Error(Exception):
    '''base class for exceptions in this module'''
    pass

class InterfaceError(Exception):
    '''raised for errors with an interface.

    this should indicate a software error, like difficulty
    connecting to an interface
    '''
    pass

class ComponentError(Exception):
    '''raised for errors with a component.

    this should indicate a hardware error in the physical world,
    like a problem with a feeder.

    this should be raised by components when doing any internal
    validation that they are working properly

    '''
    pass

class WriteCannotBeReadError(Exception):
    '''raised when an interface configured to write output cannot be read '''
    pass



import fnmatch
import os
import wave
import logging
import random
from contextlib import closing
from opyrant.utils import Event, filter_files

logger = logging.getLogger(__name__)

# TODO: Integrate this concept of "event" with the one in events.py

class Stimulus(Event):
    """docstring for Stimulus"""
    def __init__(self, *args, **kwargs):
        super(Stimulus, self).__init__(*args, **kwargs)
        if self.label=='':
            self.label = 'stimulus'


class AuditoryStimulus(Stimulus):
    """docstring for AuditoryStimulus"""
    def __init__(self, *args, **kwargs):
        super(AuditoryStimulus, self).__init__(*args, **kwargs)
        if self.label=='':
            self.label = 'auditory_stimulus'

    @classmethod
    def from_wav(cls, wavfile):

        logger.debug("Attempting to create stimulus object from %s" % wavfile)
        with closing(wave.open(wavfile,'rb')) as wf:
            (nchannels, sampwidth, framerate, nframes, comptype, compname) = wf.getparams()

            duration = float(nframes)/sampwidth
            duration = duration * 2.0 / framerate
            stim = cls(time=0.0,
                       duration=duration,
                       name=wavfile,
                       label='wav',
                       description='',
                       file_origin=wavfile,
                       annotations={'nchannels': nchannels,
                                    'sampwidth': sampwidth,
                                    'framerate': framerate,
                                    'nframes': nframes,
                                    'comptype': comptype,
                                    'compname': compname,
                                    }
                       )
        return stim


class StimulusCondition(object):
    """ Class to represent a single stimulus condition for an operant
    conditioning experiment. The name parameter should be meaningful, as it will
    be stored with the trial data. The booleans "is_rewarded" and "is_punished"
    can be used to state if a stimulus should consequated according to the
    experiment's reinforcement schedule.

    Parameters
    ----------
    name: string
        Name of the stimulus condition used in data storage
    response: string, int, or bool
        The value of the desired response. Used to determine if the subject's
        response was correct. (e.g. "left", True)
    is_rewarded: bool
        Whether or not a correct response should be rewarded
    is_punished: bool
        Whether or not an incorrect response should be punished
    files: list
        A list of files to use for the condition. If files is omitted, the list
        will be discovered using the file_path, file_pattern, and recursive
        parameters.
    file_path: string
        Path to directory where stimuli are stored
    recursive: bool
        Whether or not to search file_path recursively
    file_pattern: string
        A glob pattern to filter files by
    replacement: bool
        Whether individual stimuli should be sampled with replacement
    shuffle: bool
        Whether the list of files should be shuffled before sampling.

    Attributes
    ----------
    name: string
        Name of the stimulus condition used in data storage
    response: string, int, or bool
        The value of the desired response. Used to determine if the subject's
        response was correct. (e.g. "left", True)
    is_rewarded: bool
        Whether or not a correct response should be rewarded
    is_punished: bool
        Whether or not an incorrect response should be punished
    files: list
        All of the matching files found
    replacement: bool
        Whether individual stimuli should be sampled with replacement
    shuffle: bool
        Whether the list of files should be shuffled before sampling.

    Methods
    -------
    get()

    Examples
    --------
    # Get ".wav" files for a "go" condition of a "Go-NoGo" experiment
    condition = StimulusCondition(name="Go",
                                  response=True,
                                  is_rewarded=True,
                                  is_punished=True,
                                  file_path="/path/to/stimulus_directory",
                                  recursive=True,
                                  file_pattern="*.wav",
                                  replacement=False)

    # Get a wavefile
    wavefile = condition.get()
    """

    def __init__(self, name="", response=None, is_rewarded=True,
                 is_punished=True, files=None, file_path="", recursive=False,
                 file_pattern="*", shuffle=True, replacement=False):

        # These should do something better than printing and returning
        if files is None:
            if len(file_path) == 0:
                raise IOError("No stimulus file_path provided!")
            if not os.path.exists(file_path):
                raise IOError("Stimulus file_path does not exist! %s" % file_path)

        self.name = name
        self.response = response
        self.is_rewarded = is_rewarded
        self.is_punished = is_punished
        self.shuffle = shuffle
        self.replacement = replacement

        if files is None:
            self.files = filter_files(file_path,
                                      file_pattern=file_pattern,
                                      recursive=recursive)
        else:
            self.files = files

        self._index_list = range(len(self.files))
        if self.shuffle:
            random.shuffle(self._index_list)

        logger.debug("Created new condition: %s" % self)

    def __str__(self):

        return "".join(["Condition %s: " % self.name,
                        "# files = %d" % len(self.files)])

    def get(self):
        """ Gets a single file from this condition's list of files. If
        replacement is True, choose a file randomly with replacement. If
        replacement is False, then return files in their (possibly shuffled)
        order.
        """

        if len(self._index_list) == 0:
            self._index_list = range(len(self.files))
            if self.shuffle:
                random.shuffle(self._index_list)

        if self.replacement is True:
            index = random.choice(self._index_list)
        else:
            index = self._index_list.pop(0)

        logger.debug("Selected file %d of %d" % (index + 1, len(self.files)))
        return self.files[index]


class StimulusConditionWav(StimulusCondition):
    """ Modifies StimulusCondition to only include .wav files. For usage
    information see StimulusCondition.
    """

    def __init__(self, *args, **kwargs):

        super(StimulusConditionWav, self).__init__(file_pattern="*.wav",
                                                   *args, **kwargs)

    def get(self):
        """ Gets an AuditoryStimulus instance from a chosen .wav file """
        wavfile = super(StimulusConditionWav, self).get()

        return AuditoryStimulus.from_wav(wavfile)



## Panel classes

class BasePanel(object):
    """Returns a panel instance.

    This class should be subclassed to define a local panel configuration.

    To build a panel, do the following in the __init__() method of your local
        subclass:

    1. add instances of the necessary interfaces to the 'interfaces' dict
        attribute:
        >>> self.interfaces['comedi'] = comedi.ComediInterface(device_name='/dev/comedi0')

    2. add inputs and outputs to the 'inputs' and 'outputs' list attributes:
        >>> for in_chan in range(4):
                self.inputs.append(hwio.BooleanInput(interface=self.interfaces['comedi'],
                                                 params = {'subdevice': 2,
                                                           'channel': in_chan
                                                           },
                                                 )
    3. add components constructed from your inputs and outputs:
        >>> self.hopper = components.Hopper(IR=self.inputs[3],solenoid=self.outputs[4])

    4. assign panel methods needed for operant behavior, such as 'reward':
        >>> self.reward = self.hopper.reward

    5. finally, define a reset() method that will set the entire panel to a
        neutral state:

        >>> def reset(self):
        >>>     for output in self.outputs:
        >>>         output.set(False)
        >>>     self.house_light.write(True)
        >>>     return True

    """
    def __init__(self, *args,**kwargs):

        self.interfaces = {}

        self.inputs = []
        self.outputs = []

    def reset(self):
        """
        Turn everything off (sleep), then ready
        """
        # Should log that nothing is being reset and move on

        self.sleep()
        return self.ready()

    def ready(self):

        return True

    def reward(self):

        pass

    def sleep(self):
        """
        Turn all boolean outputs off
        """
        for output in self.outputs:
            if isinstance(output, hwio.BooleanOutput):
                self.output.write(False)

        return True

    def idle(self):

        return True

    def wake(self):
        """
        Ready the panel
        """

        return self.ready()



import logging
from opyrant.errors import WriteCannotBeReadError

logger = logging.getLogger(__name__)


class BaseIO(object):
    """ Any type of IO device. Maintains info on the interface and configuration
    params for querying the IO device

    Parameters
    ----------
    name: string
        A name given to the IO. Useful when it is used in logging and error
        messages.
    interface: subclass of base_.BaseInterface
        An instance of an interface through which writes and reads are sent
    params: dictionary
        A dictionary of parameters for configuration and write/read calls.
        Common keys are: subdevice, channel, etc.

    Attributes
    ----------
    name: string
        A name given to the IO. Useful when it is used in logging and error
        messages.
    interface: subclass of base_.BaseInterface
        An instance of an interface through which writes and reads are sent
    params: dictionary
        A dictionary of parameters for configuration and write/read calls.
        Common keys are: subdevice, channel, etc.
    """

    def __init__(self, name=None, interface=None, params={},
                 *args, **kwargs):

        super(BaseIO, self).__init__(*args, **kwargs)
        self.name = name
        self.interface = interface
        self.params = params


class BooleanInput(BaseIO):
    """ Class which holds information about boolean inputs and abstracts the
    methods of querying their values

    Parameters
    ----------
    interface: a subclass of base_.BaseInterface
        Interface through which values are read. Must have '_read_bool' method.
    params: dictionary
        A dictionary of parameters for configuration and boolean read calls.
        Common keys are: subdevice, channel, invert, etc.

    Attributes
    ----------
    name: string

    interface: a subclass of base_.BaseInterface
        Interface through which values are read. Must have '_read_bool' method.
    params: dictionary
        A dictionary of parameters for configuration and boolean read calls.
        Common keys are: subdevice, channel, invert, etc.
    last_value: bool
        Most recently returned value

    Methods
    -------
    config()
        Configures the boolean input
    read()
        Reads value of the input. Returns a boolean
    poll()
        Polls the input until value is True. Returns the time of the change
    """

    def __init__(self, interface=None, params={},
                 *args, **kwargs):
        super(BooleanInput, self).__init__(interface=interface,
                                           params=params,
                                           *args,
                                           **kwargs)

        assert self.interface.can_read_bool
        self.last_value = False
        self.config()

    def config(self):
        """ Calls the interface's _config_read method with the keyword arguments
        in params

        Returns
        -------
        bool
            True if configuration succeeded
        """

        if not hasattr(self.interface, "_config_read"):
            return False

        return self.interface._config_read(**self.params)

    def read(self):
        """ Read the status of the boolean input

        Returns
        -------
        bool
            The current status reported by the interface
        """

        self.last_value = self.interface._read_bool(**self.params)
        return self.last_value

    def poll(self, timeout=None):
        """ Runs a loop, querying for the boolean input to return True.

        Parameters
        ----------
        timeout: float

        Returns
        -------
        datetime or None
            peck time or None if timeout
        """

        input_time = self.interface._poll(timeout=timeout,
                                          last_value=self.last_value,
                                          **self.params)
        if input_time is not None:
            self.last_value = True
        else:
            self.last_value = False

        return input_time

class BooleanOutput(BaseIO):
    """Class which holds information about boolean outputs and abstracts the
    methods of writing to them

    Parameters
    ----------
    interface: subclass of base_.BaseInterface
        Interface through which values are written. Must have '_write_bool'
        method.
    params: dictionary
        A dictionary of parameters for configuration and boolean write calls.
        Common keys are: subdevice, channel, invert, etc.

    Methods
    -------
    config()
        Configures the boolean output
    write(value)
        Writes a value to the output. Returns the value
    read()
        If the interface supports '_read_bool' for this output, returns
        the current value of the output from the interface. Otherwise this
        returns the last passed by write(value)
    toggle()
        Flips the value from the current value
    """
    def __init__(self, interface=None, params={}, *args, **kwargs):
        super(BooleanOutput, self).__init__(interface=interface,
                                            params=params,
                                            *args,
                                            **kwargs)

        assert self.interface.can_write_bool
        self.last_value = None
        self.config()

    def config(self):
        """ Calls the interface's _config_write method with the keyword
        arguments in params

        Returns
        -------
        bool
            True if configuration succeeded
        """
        if not hasattr(self.interface, "_config_write"):
            return False

        logger.debug("Configuring BooleanOutput to write on interface % s" % self.interface)
        return self.interface._config_write(**self.params)

    def read(self):
        """ Read the status of the boolean output, if supported

        Returns
        -------
        bool
            The current status reported by the interface or the last value
            written to the interface.
        """
        if self.interface.can_read_bool:
            try:
                value = self.interface._read_bool(**self.params)
            except WriteCannotBeReadError:
                value = self.last_value
        else:
            value = self.last_value

        return value

    def write(self, value=False, event=None):
        """ Writes to the boolean output

        Parameters
        ----------
        value: bool
            Value to be written to the output
        event: dictionary
            Dictionary containing event details that are passed along to the
            interface.

        Returns
        -------
        bool
            The value written to the interface
        """

        logger.debug("Setting value to %s" % value)
        self.last_value = self.interface._write_bool(value=value,
                                                     event=event,
                                                     **self.params)
        return self.last_value

    def toggle(self, event=None):
        """ Toggles the value of the boolean output

        Parameters
        ----------
        event: dictionary
            Dictionary containing event details that are passed along to the
            interface.

        Returns
        -------
        bool
            The value written to the interface
        """

        # TODO: what will event be here?
        value = not self.read()
        return self.write(value=value, event=event)


class AnalogInput(BaseIO):
    """ Class which holds information about analog inputs and abstracts the
    methods of reading from them

    Parameters
    ----------
    interface: subclass of base_.BaseInterface
        Interface through which values are written. Must have '_read_analog'
        method.
    params: dictionary
        A dictionary of parameters for configuration and analog read calls.
        Common keys are: subdevice, channel, etc.

    Methods
    -------
    config()
        Configures the analog input
    read(nsamples)
        Reads nsamples values from the input. Returns the values as an array
    """
    def __init__(self, interface=None, params={}, *args, **kwargs):
        super(AnalogInput, self).__init__(interface=interface,
                                          params=params,
                                          *args,
                                          **kwargs)
        assert self.interface.can_read_analog

        self.config()

    def config(self):
        """ Calls the interface's _config_read_analog method with the keyword
        arguments in params

        Returns
        -------
        bool
            True if configuration succeeded
        """
        if not hasattr(self.interface, "_config_read_analog"):
            return False

        logger.debug("Configuring AnalogInput to read on interface % s" % self.interface)
        return self.interface._config_read_analog(**self.params)

    def read(self, nsamples):
        """ Reads from the analog input

        Parameters
        ----------
        nsamples: int
            Number of samples to read from input

        Returns
        -------
        numpy array
            The analog signal recorded by the interface
        """
        return self.interface._read_analog(nsamples=nsamples, **self.params)


class AnalogOutput(BaseIO):
    """ Class which holds information about analog outputs and abstracts the
    methods of writing to them

    Parameters
    ----------
    interface: subclass of base_.BaseInterface
        Interface through which values are written. Must have '_write_analog'
        method.
    params: dictionary
        A dictionary of parameters for configuration and analog write calls.
        Common keys are: subdevice, channel, etc.

    Methods
    -------
    config()
        Configures the analog output
    write(values)
        Writes an array of values to the output. Returns True if successful.
    """
    def __init__(self, interface=None, params={}, *args, **kwargs):
        super(AnalogOutput, self).__init__(interface=interface,
                                           params=params,
                                           *args,
                                           **kwargs)
        assert self.interface.can_write_analog

        self.config()

    def config(self):
        """ Calls the interface's config_write_analog method with the keyword
        arguments in params

        Returns
        -------
        bool
            True if configuration succeeded
        """
        if not hasattr(self.interface, "_config_write_analog"):
            return False

        logger.debug("Configuring AnalogOutput to write on interface % s" % self.interface)
        return self.interface._config_write_analog(**self.params)

    def write(self, values, event=None):
        """ Writes to the analog output

        Parameters
        ----------
        values: numpy array
            Array of float values to be written to the output
        event: dictionary
            Dictionary containing event details that are passed along to the
            interface.

        Returns
        -------
        bool
            True if the write succeeded
        """

        return self.interface._write_analog(values=values, event=event)


class AudioOutput(BaseIO):
    """ Class which holds information about audio outputs and abstracts the
    methods of writing to them

    Parameters
    ----------
    interface: subclass of base_.BaseInterface()
        Must have the methods '_queue_wav', '_play_wav', '_stop_wav'
    params: dictionary
        A dictionary of parameters for configuration and audio playback calls.
        Common keys are: subdevice, channel, etc.

    Methods:
    config()
        Configures the audio interface
    queue(wav_filename)
        Queues a .wav file for playback
    play()
        Plays the queued audio file
    stop()
        Stops the playing audio file
    """

    def __init__(self, interface=None, params={}, *args, **kwargs):
        super(AudioOutput, self).__init__(interface=interface,
                                          params=params,
                                          *args,
                                          **kwargs)

        assert hasattr(self.interface, '_queue_wav')
        assert hasattr(self.interface, '_play_wav')
        assert hasattr(self.interface, '_stop_wav')
        self.config()

    def config(self):
        """ Calls the interface's config_write_analog method with the keyword
        arguments in params

        Returns
        -------
        bool
            True if configuration succeeded
        """
        if not hasattr(self.interface, "_config_write_analog"):
            return False

        logger.debug("Configuring AudioOutput to write on interface % s" % self.interface)
        return self.interface._config_write_analog(**self.params)

    def queue(self, wav_filename, event=None):
        return self.interface._queue_wav(wav_filename, event=event, **self.params)

    def play(self, event=None):
        return self.interface._play_wav(event=event, **self.params)

    def stop(self, event=None):
        return self.interface._stop_wav(event=event, **self.params)



from opyrant import hwio, components, panels, utils
from opyrant.interfaces import comedi_, pyaudio_

_VOGEL_MAP = {
    1: ('/dev/comedi0', 2, 0, 2, 8), # box_id:(subdevice,in_dev,in_chan,out_dev,out_chan)
    2: ('/dev/comedi0', 2, 4, 2, 16),
    3: ('/dev/comedi0', 2, 24, 2, 32),
    4: ('/dev/comedi0', 2, 28, 2, 40),
    5: ('/dev/comedi0', 2, 48, 2, 56),
    6: ('/dev/comedi0', 2, 52, 2, 64),
    7: ('/dev/comedi0', 2, 72, 2, 80),
    8: ('/dev/comedi0', 2, 76, 2, 88),
    }

class VogelPanel(panels.BasePanel):
    """class for vogel boxes """
    def __init__(self,id=None, *args, **kwargs):
        super(VogelPanel, self).__init__(*args, **kwargs)
        self.id = id

        # define interfaces
        self.interfaces['comedi'] = comedi_.ComediInterface(device_name=_VOGEL_MAP[self.id][0])
        self.interfaces['pyaudio'] = pyaudio_.PyAudioInterface(device_name='dac%i'%self.id)


        # define inputs
        for in_chan in [ii+_VOGEL_MAP[self.id][2] for ii in range(4)]:
            self.inputs.append(hwio.BooleanInput(interface=self.interfaces['comedi'],
                                                 params = {'subdevice': _VOGEL_MAP[self.id][1],
                                                           'channel': in_chan
                                                           },
                                                 )
                               )
        for out_chan in [ii+_VOGEL_MAP[self.id][4] for ii in range(8)]:
            self.outputs.append(hwio.BooleanOutput(interface=self.interfaces['comedi'],
                                                 params = {'subdevice': _VOGEL_MAP[self.id][3],
                                                           'channel': out_chan
                                                           },
                                                   )
                                )
        self.speaker = hwio.AudioOutput(interface=self.interfaces['pyaudio'])

        # assemble inputs into components
        self.left = components.PeckPort(IR=self.inputs[0],LED=self.outputs[0],name='l')
        self.center = components.PeckPort(IR=self.inputs[1],LED=self.outputs[1],name='c')
        self.right = components.PeckPort(IR=self.inputs[2],LED=self.outputs[2],name='r')
        self.house_light = components.HouseLight(light=self.outputs[3])
        self.hopper = components.Hopper(IR=self.inputs[3],solenoid=self.outputs[4])

        # define reward & punishment methods
        self.reward = self.hopper.reward
        self.punish = self.house_light.punish

    def reset(self):
        for output in self.outputs:
            output.write(False)
        self.house_light.on()
        self.hopper.down()

    def test(self):
        print ('reset')
        self.reset()
        dur = 2.0
        for output in self.outputs:
            print ('output %s on' % output)
            output.write(True)
            utils.wait(dur)
            print ('output %s off' % output)
            output.write(False)
        print ('reset')
        self.reset()
        print ('feed')
        self.reward(value=dur)
        print ('timeout')
        self.punish(value=dur)
        print ('queue file')
        self.speaker.queue('/usr/local/stimuli/A1.wav')
        print ('play file')
        self.speaker.play()
        return True


class Vogel1(VogelPanel):
    """Vogel1 panel"""
    def __init__(self):
        super(Vogel1, self).__init__(id=1)

class Vogel2(VogelPanel):
    """Vogel2 panel"""
    def __init__(self):
        super(Vogel2, self).__init__(id=2)

class Vogel3(VogelPanel):
    """Vogel3 panel"""
    def __init__(self):
        super(Vogel3, self).__init__(id=3)

class Vogel4(VogelPanel):
    """Vogel4 panel"""
    def __init__(self):
        super(Vogel4, self).__init__(id=4)

class Vogel6(VogelPanel):
    """Vogel6 panel"""
    def __init__(self):
        super(Vogel6, self).__init__(id=6)

class Vogel7(VogelPanel):
    """Vogel7 panel"""
    def __init__(self):
        super(Vogel7, self).__init__(id=7)

class Vogel8(VogelPanel):
    """Vogel8 panel"""
    def __init__(self):
        super(Vogel8, self).__init__(id=8)


# in the end, 'PANELS' should contain each operant panel available for use

PANELS = {"1": Vogel1,
          "2": Vogel2,
          "3": Vogel3,
          "4": Vogel4,
          "6": Vogel6,
          "7": Vogel7,
          "8": Vogel8,
          }

BEHAVIORS = ['opyrant.behavior',
             'glab_behaviors',
            ]

DATA_PATH = '/home/bird/opdat/'

# SMTP_CONFIG

DEFAULT_EMAIL = 'justin.kiggins@gmail.com'

SMTP_CONFIG = {'mailhost': 'localhost',
               'toaddrs': [DEFAULT_EMAIL],
               'fromaddr': 'Vogel <bird@vogel.ucsd.edu>',
               'subject': '[opyrant notice] on vogel',
               'credentials': None,
               'secure': None,
               }



from opyrant import hwio, components, panels, utils
from opyrant.interfaces import comedi_, pyaudio_
from opyrant import InterfaceError
import time

_ZOG_MAP = {
    1: ('/dev/comedi0', 2, 0, 2, 8), # box_id:(subdevice,in_dev,in_chan,out_dev,out_chan)
    2: ('/dev/comedi0', 2, 4, 2, 16),
    3: ('/dev/comedi0', 2, 24, 2, 32),
    4: ('/dev/comedi0', 2, 28, 2, 40),
    5: ('/dev/comedi0', 2, 48, 2, 56),
    6: ('/dev/comedi0', 2, 52, 2, 64),
    7: ('/dev/comedi0', 2, 72, 2, 80),
    8: ('/dev/comedi0', 2, 76, 2, 88),
    9: ('/dev/comedi1', 2, 0, 2, 8),
    10: ('/dev/comedi1', 2, 4, 2, 16),
    11: ('/dev/comedi1', 2, 24, 2, 32),
    12: ('/dev/comedi1', 2, 28, 2, 40),
    13: ('/dev/comedi1', 2, 48, 2, 56),
    14: ('/dev/comedi1', 2, 52, 2, 64),
    15: ('/dev/comedi1', 2, 72, 2, 80),
    16: ('/dev/comedi1', 2, 76, 2, 88),
    }

dev_name_fmt = 'Adapter 1 (5316) - Output Stream %i'

class ZogAudioInterface(pyaudio_.PyAudioInterface):
    """docstring for ZogAudioInterface"""
    def __init__(self, *args, **kwargs):
        super(ZogAudioInterface, self).__init__(*args,**kwargs)
    def validate(self):
        super(ZogAudioInterface, self).validate()
        if self.wf.getframerate()==48000:
            return True
        else:
            raise InterfaceError('this wav file must be 48kHz')


class ZogPanel(panels.BasePanel):
    """class for zog boxes """
    def __init__(self,id=None, *args, **kwargs):
        super(ZogPanel, self).__init__(*args, **kwargs)
        self.id = id

        # define interfaces
        self.interfaces['comedi'] = comedi_.ComediInterface(device_name=_ZOG_MAP[self.id][0])
        self.interfaces['pyaudio'] = ZogAudioInterface(device_name= (dev_name_fmt % self.id))


        # define inputs
        for in_chan in [ii+_ZOG_MAP[self.id][2] for ii in range(4)]:
            self.inputs.append(hwio.BooleanInput(interface=self.interfaces['comedi'],
                                                 params = {'subdevice': _ZOG_MAP[self.id][1],
                                                           'channel': in_chan
                                                           },
                                                 )
                               )
        for out_chan in [ii+_ZOG_MAP[self.id][4] for ii in range(8)]:
            self.outputs.append(hwio.BooleanOutput(interface=self.interfaces['comedi'],
                                                 params = {'subdevice': _ZOG_MAP[self.id][3],
                                                           'channel': out_chan
                                                           },
                                                   )
                                )
        self.speaker = hwio.AudioOutput(interface=self.interfaces['pyaudio'])

        # assemble inputs into components
        self.left = components.PeckPort(IR=self.inputs[0],LED=self.outputs[0],name='l')
        self.center = components.PeckPort(IR=self.inputs[1],LED=self.outputs[1],name='c')
        self.right = components.PeckPort(IR=self.inputs[2],LED=self.outputs[2],name='r')
        self.house_light = components.HouseLight(light=self.outputs[3])
        self.hopper = components.Hopper(IR=self.inputs[3],solenoid=self.outputs[4])

        # define reward & punishment methods
        self.reward = self.hopper.reward
        self.punish = self.house_light.punish

    def reset(self):
        for output in self.outputs:
            output.write(False)
        self.house_light.on()
        self.hopper.down()
        # self.speaker.stop()

    def test(self):
        self.reset()
        dur = 2.0
        for output in self.outputs:
            output.write(True)
            utils.wait(dur)
            output.write(False)
        self.reset()
        self.reward(value=dur)
        self.punish(value=dur)
        self.speaker.queue('/usr/local/stimuli/test48k.wav')
        self.speaker.play()
        time.sleep(1.0)
        self.speaker.stop()
        return True



class Zog1(ZogPanel):
    """Zog1 panel"""
    def __init__(self):
        super(Zog1, self).__init__(id=1)

class Zog2(ZogPanel):
    """Zog2 panel"""
    def __init__(self):
        super(Zog2, self).__init__(id=2)

class Zog3(ZogPanel):
    """Zog3 panel"""
    def __init__(self):
        super(Zog3, self).__init__(id=3)

class Zog4(ZogPanel):
    """Zog4 panel"""
    def __init__(self):
        super(Zog4, self).__init__(id=4)

class Zog6(ZogPanel):
    """Zog6 panel"""
    def __init__(self):
        super(Zog6, self).__init__(id=6)

class Zog8(ZogPanel):
    """Zog8 panel"""
    def __init__(self):
        super(Zog8, self).__init__(id=8)

class Zog10(ZogPanel):
    """Zog10 panel"""
    def __init__(self):
        super(Zog10, self).__init__(id=10)

class Zog12(ZogPanel):
    """Zog12 panel"""
    def __init__(self):
        super(Zog12, self).__init__(id=12)

class Zog13(ZogPanel):
    """Zog13 panel"""
    def __init__(self):
        super(Zog13, self).__init__(id=13)

class Zog14(ZogPanel):
    """Zog14 panel"""
    def __init__(self):
        super(Zog14, self).__init__(id=14)

class Zog15(ZogPanel):
    """Zog15 panel"""
    def __init__(self):
        super(Zog15, self).__init__(id=15)

class Zog16(ZogPanel):
    """Zog16 panel"""
    def __init__(self):
        super(Zog16, self).__init__(id=16)


# define the panels with cue lights
class ZogCuePanel(ZogPanel):
    """ZogCuePanel panel"""
    def __init__(self,id=None):
        super(ZogCuePanel, self).__init__(id=id)

        for out_chan in [ii+_ZOG_MAP[self.id][4] for ii in range(5,8)]:
            self.outputs.append(hwio.BooleanOutput(interface=self.interfaces['comedi'],
                                                 params = {'subdevice': _ZOG_MAP[self.id][3],
                                                           'channel': out_chan
                                                           },
                                                   )
                                )
        self.cue = components.RGBLight(red=self.outputs[7],
                                       green=self.outputs[5],
                                       blue=self.outputs[6],
                                       name='cue')

class Zog5(ZogCuePanel):
    """Zog5 panel"""
    def __init__(self):
        super(Zog5, self).__init__(id=5)

class Zog7(ZogCuePanel):
    """Zog7 panel"""
    def __init__(self):
        super(Zog7, self).__init__(id=7)

class Zog9(ZogCuePanel):
    """Zog9 panel"""
    def __init__(self):
        super(Zog9, self).__init__(id=9)

class Zog11(ZogCuePanel):
    """Zog11 panel"""
    def __init__(self):
        super(Zog11, self).__init__(id=11)

class Zog10(ZogCuePanel):
    """Zog10 panel"""
    def __init__(self):
        super(Zog10, self).__init__(id=10)

class Zog12(ZogCuePanel):
    """Zog12 panel"""
    def __init__(self):
        super(Zog12, self).__init__(id=12)


# in the end, 'PANELS' should contain each operant panel available for use

PANELS = {
          "1": Zog1,
          "2": Zog2,
          "3": Zog3,
          "4": Zog4,
          "5": Zog5,
          "6": Zog6,
          "7": Zog7,
          "8": Zog8,
          "9": Zog9,
          "10": Zog10,
          "11": Zog11,
          "12": Zog12,
          "13": Zog13,
          "14": Zog14,
          "15": Zog15,
          "16": Zog16,
          }

BEHAVIORS = ['opyrant.behavior',
             'glab_behaviors'
            ]

DATA_PATH = '/home/bird/opdat/'

# SMTP_CONFIG

DEFAULT_EMAIL = 'justin.kiggins@gmail.com'

SMTP_CONFIG = {'mailhost': 'localhost',
               'toaddrs': [DEFAULT_EMAIL],
               'fromaddr': 'Zog <bird@zog.ucsd.edu>',
               'subject': '[opyrant notice] on zog',
               'credentials': None,
               'secure': None,
               }



import logging
from opyrant import queues, reinf, utils, trials

logger = logging.getLogger(__name__)


class Block(queues.BaseHandler):
    """ Class that allows one to iterate over a block of trials according to a
    specific queue.

    Parameters
    ----------
    conditions: list
        A list of StimulusConditions to iterate over according to the queue
    index: int
        Index of the block
    experiment: instance of Experiment class
        The experiment of which this block is a part
    queue: a queue function or Class
        The queue used to iterate over trials for this block
    reinforcement: instance of Reinforcement class (ContinuousReinforcement())
        The reinforcement schedule to use for this block.
    Additional key-value pairs are used to initialize the trial queue

    Attributes
    ----------
    conditions: list
        A list of StimulusConditions to iterate over according to the queue
    index: int
        Index of the block
    experiment: instance of Experiment class
        The experiment of which this block is a part
    queue: queue generator or class instance
        The queue that will be iterated over.
    reinforcement: instance of Reinforcement class (ContinuousReinforcement())
        The reinforcement schedule to use for this block.

    Examples
    --------
    # Initialize a block with a random queue, and at most 200 trials.
    trials = Block(conditions,
                   experiment=e,
                   queue=queues.random_queue,
                   max_items=200)
    for trial in trials:
        trial.run()
    """

    def __init__(self, conditions, index=0, experiment=None,
                 queue=queues.random_queue, reinforcement=None,
                 **queue_parameters):

        if conditions is None:
            raise ValueError("Block must be called with a list of conditions")

        # Could check to ensure reinforcement is of the correct type
        if reinforcement is None:
            reinforcement = reinf.ContinuousReinforcement()

        super(Block, self).__init__(queue=queue,
                                    items=conditions,
                                    **queue_parameters)

        self.index = index
        self.experiment = experiment
        self.conditions = conditions
        self.reinforcement = reinforcement

        logger.debug("Initialize block: %s" % self)

    def __str__(self):

        desc = ["Block"]
        if self.conditions is not None:
            desc.append("%d stimulus conditions" % len(self.conditions))
        if self.queue is not None:
            desc.append("queue = %s" % self.queue.__name__)

        return " - ".join(desc)

    def check_completion(self):

        # if self.end is not None:
        #     if utils.check_time((self.start, self.end)): # Will start ever be none? Shouldn't be.
        #         logger.debug("Block is complete due to time")
        #         return True # Block is complete

        # if self.max_trials is not None:
        #     if self.num_trials >= self.max_trials:
        #         logger.debug("Block is complete due to trial count")
        #         return True

        return False

    def __iter__(self):

        # Loop through the queue generator
        trial_index = 0
        for condition in self.queue:
            # Create a trial instance
            trial_index += 1
            trial = trials.Trial(index=trial_index,
                                 experiment=self.experiment,
                                 condition=condition,
                                 block=self)
            yield trial


class BlockHandler(queues.BaseHandler):
    """ Class which enables iterating over blocks of trials

    Parameters
    ----------
    blocks: list
        A list of Block objects
    queue: a queue function or Class
        The queue used to iterate over blocks
    Additional key-value pairs are used to initialize the queue

    Attributes
    ----------
    block_index: int
        Index of the current block
    blocks: list
        A list of Block objects
    queue: queue generator or class instance
        The queue that will be iterated over.
    queue_parameters: dict
        All additional parameters used to initialize the queue.

    Example
    -------
    # Initialize the BlockHandler
    blocks = BlockHandler(blocks, queue=queues.block_queue)
    # Loop through the blocks, then loop through all trials in the block
    for block in blocks:
        for trial in block:
            trial.run()
    """

    def __init__(self, blocks, queue=queues.block_queue, **queue_parameters):

        self.blocks = blocks
        self.block_index = 0
        super(BlockHandler, self).__init__(queue=queue,
                                           items=blocks,
                                           **queue_parameters)

    def __iter__(self):

        for block in self.queue:
            self.block_index += 1
            block.index = self.block_index
            yield block



import datetime
from opyrant import hwio, utils, ComponentError

class BaseComponent(object):
    """Base class for physcal component

    Parameters
    ----------
    name: string
        The name of the component to be stored in events (defaults to the
        class name)

    Attributes
    ----------
    name: string
        The name of the component to be stored in events (defaults to the
        class name)
    event:
        A dictionary that is sent along with changes to the state of inputs and
        outputs. It contains three main keys: name, action, and metadata. name
        is set here, action is set when a component method is called, and
        metadata is set as an optional argument to some methods.
    """
    def __init__(self, name=None, *args, **kwargs):
        if name is None:
            name = self.__class__.__name__
        self.name = name
        self.event = dict(name=self.name,
                          action="",
                          metadata=None)
        pass


## Hopper ##
class HopperActiveError(ComponentError):
    """raised when the hopper is up when it shouldn't be"""
    pass


class HopperInactiveError(ComponentError):
    """raised when the hopper is down when it shouldn't be"""
    pass


class HopperAlreadyUpError(HopperActiveError):
    """raised when the hopper is already up before it goes up"""
    pass


class HopperWontComeUpError(HopperInactiveError):
    """raised when the hopper won't come up"""
    pass


class HopperWontDropError(HopperActiveError):
    """raised when the hopper won't drop"""
    pass


class Hopper(BaseComponent):
    """ Class which holds information about a hopper

    Parameters
    ----------
    solenoid : `hwio.BooleanOutput`
        output channel to activate the solenoid & raise the hopper
    IR : :class:`hwio.BooleanInput`
       input channel for the IR beam to check if the hopper is up (optional)
    max_lag : float, optional
        time in seconds to wait before checking to make sure the hopper is up
        (default=0.3)

    Attributes
    ----------
    solenoid : hwio.BooleanOutput
        output channel to activate the solenoid & raise the hopper
    IR : hwio.BooleanInput
       input channel for the IR beam to check if the hopper is up (optional)
    max_lag : float
        time in seconds to wait before checking to make sure the hopper is up

    """
    def __init__(self, solenoid, IR=None, max_lag=0.3, *args, **kwargs):
        super(Hopper, self).__init__(*args, **kwargs)
        self.max_lag = max_lag

        if (IR is not None) and (not isinstance(IR, hwio.BooleanInput)):
            raise ValueError('%s is not an input channel' % IR)
        self.IR = IR

        if isinstance(solenoid, hwio.BooleanOutput):
            self.solenoid = solenoid
        else:
            raise ValueError('%s is not an output channel' % solenoid)

    def check(self):
        """ Reads the status of solenoid & IR beam, then throws an error if they
        don't match. If IR is None, then trust the solenoid's status.

        Returns
        -------
        bool
            True if the hopper is up.

        Raises
        ------
        HopperActiveError
            The Hopper is up and it shouldn't be. (The IR beam is tripped, but
            the solenoid is not active.)
        HopperInactiveError
            The Hopper is down and it shouldn't be. (The IR beam is not tripped,
            but the solenoid is active.)

        """
        if self.IR is None:
            return self.solenoid.read() is True

        IR_status = self.IR.read()
        solenoid_status = self.solenoid.read()
        if IR_status != solenoid_status:
            if IR_status:
                raise HopperActiveError
            elif solenoid_status:
                raise HopperInactiveError
            else:
                raise ComponentError("the IR & solenoid don't match: IR:%s,solenoid:%s" % (IR_status, solenoid_status))
        else:
            return IR_status

    def up(self):
        """ Raises the hopper up.

        Returns
        -------
        datetime
            Time at which the hopper came up

        Raises
        ------
        HopperWontComeUpError
            The Hopper did not raise.
        """

        self.event["action"] = "up"
        self.solenoid.write(True, event=self.event)
        if self.IR is None:
            return datetime.datetime.now()

        time_up = self.IR.poll(timeout=self.max_lag)

        if time_up is None:  # poll timed out
            self.solenoid.write(False)
            raise HopperWontComeUpError
        else:
            return time_up

    def down(self):
        """ Lowers the hopper.

        Returns
        -------
        datetime
            Time at which the hopper came down

        Raises
        ------
        HopperWontDropError
            The Hopper did not drop.
        """
        self.event["action"] = "down"
        self.solenoid.write(False, event=self.event)
        time_down = datetime.datetime.now()
        utils.wait(self.max_lag)
        try:
            self.check()
        except HopperActiveError as e:
            raise HopperWontDropError(e)
        return time_down

    def feed(self, dur=2.0, error_check=True):
        """ Performs a feed

        Parameters
        ----------
        dur : float, optional
            duration of feed in seconds

        Returns
        -------
        (datetime, float)
            Timestamp of the feed and the feed duration


        Raises
        ------
        HopperAlreadyUpError
            The Hopper was already up at the beginning of the feed.
        HopperWontComeUpError
            The Hopper did not raise for the feed.
        HopperWontDropError
            The Hopper did not drop fater the feed.

        """
        assert self.max_lag < dur, "max_lag (%ss) must be shorter than duration (%ss)" % (self.max_lag, dur)
        try:
            self.check()
        except HopperActiveError as e:
            self.solenoid.write(False)
            raise HopperAlreadyUpError(e)
        feed_time = self.up()
        utils.wait(dur)
        feed_over = self.down()
        feed_duration = feed_over - feed_time
        return (feed_time, feed_duration)

    def reward(self, value=2.0):
        """ Performs a feed as a reward

        Parameters
        ----------
        value : float, optional
            duration of feed in seconds

        Returns
        -------
        (datetime, float)
            Timestamp of the feed and the feed duration
        """
        return self.feed(dur=value)


class Button(BaseComponent):
    """ Class which holds information about buttons with an input but no output.
    Could also describe a perch.

    Parameters
    ----------
    IR : hwio.BooleanInput
        input channel for the IR beam to check for a peck

    Attributes
    ----------
    IR : hwio.BooleanInput
        input channel for the IR beam to check for a peck
    """
    def __init__(self, IR, *args, **kwargs):
        super(Button, self).__init__(*args, **kwargs)
        if isinstance(IR, hwio.BooleanInput):
            self.IR = IR
        else:
            raise ValueError('%s is not an input channel' % IR)

    def status(self):
        """ Reads the status of the IR beam

        Returns
        -------
        bool
            True if beam is broken
        """
        return self.IR.read()

    def poll(self, timeout=None):
        """ Polls the peck port until there is a peck

        Returns
        -------
        datetime
            Timestamp of the IR beam being broken.
        """
        return self.IR.poll(timeout=timeout)


## Peck Port ##
class PeckPort(BaseComponent):
    """ Class which holds information about peck ports

    Parameters
    ----------
    LED : hwio.BooleanOutput
        output channel to activate the LED in the peck port
    IR : hwio.BooleanInput
        input channel for the IR beam to check for a peck

    Attributes
    ----------
    LED : hwio.BooleanOutput
        output channel to activate the LED in the peck port
    IR : hwio.BooleanInput
        input channel for the IR beam to check for a peck

    """
    def __init__(self, IR, LED, *args, **kwargs):
        super(PeckPort, self).__init__(*args, **kwargs)
        if isinstance(IR, hwio.BooleanInput):
            self.IR = IR
        else:
            raise ValueError('%s is not an input channel' % IR)

        if isinstance(LED, hwio.BooleanOutput):
            self.LED = LED
        else:
            raise ValueError('%s is not an output channel' % LED)

    def status(self):
        """ Reads the status of the IR beam

        Returns
        -------
        bool
            True if beam is broken
        """
        return self.IR.read()

    def off(self):
        """ Turns the LED off

        Returns
        -------
        bool
            True if successful
        """
        self.event["action"] = "off"
        self.LED.write(False, event=self.event)
        return True

    def on(self):
        """ Turns the LED on

        Returns
        -------
        bool
            True if successful
        """
        self.event["action"] = "on"
        self.LED.write(True, event=self.event)
        return True

    def flash(self, dur=1.0, isi=0.1):
        """ Flashes the LED on and off with *isi* seconds high and low for *dur*
        seconds, then revert LED to prior state.

        Parameters
        ----------
        dur : float, optional
            Duration of the light flash in seconds.
        isi : float,optional
            Time interval between toggles. (0.5 * period)

        Returns
        -------
        (datetime, float)
            Timestamp of the flash and the flash duration
        """
        LED_state = self.LED.read()
        flash_time = datetime.datetime.now()
        flash_duration = datetime.datetime.now() - flash_time
        while flash_duration < datetime.timedelta(seconds=dur):
            self.LED.toggle()
            utils.wait(isi)
            flash_duration = datetime.datetime.now() - flash_time
        self.LED.write(LED_state)
        return (flash_time, flash_duration)

    def poll(self, timeout=None):
        """ Polls the peck port until there is a peck

        Returns
        -------
        datetime
            Timestamp of the IR beam being broken.
        """
        return self.IR.poll(timeout=timeout)


## House Light ##
class HouseLight(BaseComponent):
    """ Class which holds information about the house light

    Parameters
    ----------
    light : hwio.BooleanOutput
        output channel to turn the light on and off

    Attributes
    ----------
    light : hwio.BooleanOutput
        output channel to turn the light on and off

    """
    def __init__(self, light, *args, **kwargs):
        super(HouseLight, self).__init__(*args, **kwargs)
        if isinstance(light, hwio.BooleanOutput):
            self.light = light
        else:
            raise ValueError('%s is not an output channel' % light)

    def off(self):
        """Turns the house light off.

        Returns
        -------
        bool
            True if successful.

        """
        self.event["action"] = "off"
        self.light.write(False, event=self.event)
        return True

    def on(self):
        """Turns the house light on.

        Returns
        -------
        bool
            True if successful.
        """
        self.event["action"] = "on"
        self.light.write(True, event=self.event)
        return True

    def timeout(self, dur=10.0):
        """Turn off the light for *dur* seconds

        Parameters
        ----------
        dur : float, optional
            The amount of time (in seconds) to turn off the light.

        Returns
        -------
        (datetime, float)
            Timestamp of the timeout and the timeout duration

        """
        timeout_time = datetime.datetime.now()
        self.off()
        utils.wait(dur)
        timeout_duration = datetime.datetime.now() - timeout_time
        self.on()
        return (timeout_time, timeout_duration)

    def punish(self, value=10.0):
        """ Turns light off as a punishment

        Parameters
        ---------
        value : float, optional
            duration of timeout in seconds

        Returns
        -------
        (datetime, float)
            Timestamp of the timeout and the timeout duration
        """
        return self.timeout(dur=value)


## Cue Light ##
class RGBLight(BaseComponent):
    """ Class which holds information about an RGB cue light

    Parameters
    ----------
    red : hwio.BooleanOutput
        output channel for the red LED
    green : hwio.BooleanOutput
        output channel for the green LED
    blue : hwio.BooleanOutput
        output channel for the blue LED

    Attributes
    ----------
    _red : hwio.BooleanOutput
        output channel for the red LED
    _green : hwio.BooleanOutput
        output channel for the green LED
    _blue : hwio.BooleanOutput
        output channel for the blue LED

    """
    def __init__(self, red, green, blue, *args, **kwargs):
        super(RGBLight, self).__init__(*args,**kwargs)
        if isinstance(red, hwio.BooleanOutput):
            self._red = red
        else:
            raise ValueError('%s is not an output channel' % red)

        if isinstance(green, hwio.BooleanOutput):
            self._green = green
        else:
            raise ValueError('%s is not an output channel' % green)

        if isinstance(blue, hwio.BooleanOutput):
            self._blue = blue
        else:
            raise ValueError('%s is not an output channel' % blue)

    def red(self):
        """Turns the cue light to red

        Returns
        -------
        bool
            `True` if successful.
        """
        self.event["action"] = "red"
        self._green.write(False)
        self._blue.write(False)
        return self._red.write(True, event=self.event)

    def green(self):
        """Turns the cue light to green

        Returns
        -------
        bool
            `True` if successful.
        """
        self.event["action"] = "green"
        self._red.write(False)
        self._blue.write(False)
        return self._green.write(True, event=self.event)

    def blue(self):
        """Turns the cue light to blue

        Returns
        -------
        bool
            `True` if successful.
        """
        self.event["action"] = "blue"
        self._red.write(False)
        self._green.write(False)
        return self._blue.write(True, event=self.event)

    def off(self):
        """Turns the cue light off

        Returns
        -------
        bool
            `True` if successful.
        """
        self.event["action"] = "off"
        self._red.write(False, event=self.event)
        self._green.write(False)
        self._blue.write(False)
        return True


class Speaker(BaseComponent):
    """ Class which holds information about a speaker

    Parameters
    ----------
    output: hwio.AudioOutput
        Output of the speaker

    Attributes
    ----------
    output: hwio.AudioOutput
        Output of the speaker
    """

    def __init__(self, output, *args, **kwargs):

        super(Speaker, self).__init__(*args, **kwargs)
        self.output = output

    def queue(self, wav_filename, metadata=None):

        self.event["action"] = "queue"
        self.event["metadata"] = metadata
        return self.output.queue(wav_filename, event=self.event)

    def play(self):

        self.event["action"] = "play"
        return self.output.play(event=self.event)

    def stop(self):

        self.event["action"] = "stop"
        return self.output.stop(event=self.event)


class Microphone(BaseComponent):
    """ Class which holds information about a microphone

    Parameters
    ----------
    input_: hwio.AnalogInput
        Input to the microphone

    Attributes
    ----------
    input: hwio.AnalogInput
        Input to the microphone
    """

    def __init__(self, input_, *args, **kwargs):

        super(Speaker, self).__init__(*args, **kwargs)
        self.input = input_

    def record(self, nsamples):
        """ Reads from input for a set number of samples

        Parameters
        ----------
        nsamples: int
            Number of samples to read from input

        Returns
        -------
        numpy array
            The analog signal recorded by input
        """
        # TODO: This should use a stop signal too, I think
        self.event["action"] = "rec"
        return self.input.read(nsamples, event=self.event)

# ## Perch ##

# class Perch(BaseComponent):
#     """Class which holds information about a perch

#     Has parts:
#     - IR Beam (input)
#     - speaker
#     """
#     def __init__(self,*args,**kwargs):
#         super(Perch, self).__init__(*args,**kwargs)



import os


class ConfigureJSON(object):

    @classmethod
    def load(cls, config_file):
        """ Load experiment parameters from a JSON configuration file

        Parameters
        ----------
        config_file: string
            path to a JSON configuration file

        Returns
        -------
        dictionary (or list of dictionaries) of parameters to pass to a behavior
        """
        try:
            import simplejson as json
        except ImportError:
            import json

        with open(config_file, 'rb') as config:
            parameters = json.load(config)

        return parameters

    @staticmethod
    def save(parameters, filename, overwrite=False):
        """ Save a dictionary of parameters to an experiment JSON config file

        Parameters
        ----------
        parameters: dictionary
            experiment parameters
        filename: string
            path to output file
        overwrite: bool
            whether or not to overwrite if the output file already exists
        """
        try:
            import simplejson as json
        except ImportError:
            import json

        if os.path.exists(filename) and (overwrite is False):
            raise IOError("File %s already exists! To overwrite, set overwrite=True" % filename)

        with open(filename, "w") as json_file:
            json.dump(parameters,
                      json_file,
                      sort_keys=True,
                      indent=4,
                      separators=(",", ":"))


class ConfigureYAML(object):
    """ Configuration using YAML files. Thanks to pyyaml (http://pyyaml.org/wiki/PyYAMLDocumentation), this type of configuration file can be very flexible. It supports multiple experiments in the form of multiple documents in one file. It also allows for expressing native python objects (including custom classes) directly in the configuration file.
    """

    @classmethod
    def load(cls, config_file):
        """ Load experiment parameters from a YAML configuration file

        Parameters
        ----------
        config_file: string
            path to a YAML configuration file

        Returns
        -------
        dictionary (or list of dictionaries) of parameters to pass to a behavior
        """
        try:
            import yaml
        except ImportError:
            raise ImportError("Pyyaml is required to use a .yaml configuration file")

        parameters = list()
        with open(config_file, "rb") as config:
            for val in yaml.load_all(config):
                parameters.append(val)

        if len(parameters) == 1:
            parameters = parameters[0]

        return parameters

    @staticmethod
    def save(parameters, filename, overwrite=False):
        """ Save a dictionary of parameters to an experiment YAML config file

        Parameters
        ----------
        parameters: dictionary
            experiment parameters
        filename: string
            path to output file
        overwrite: bool
            whether or not to overwrite if the output file already exists
        """
        try:
            import yaml
        except ImportError:
            raise ImportError("Pyyaml is required to use a .yaml configuration file")

        if os.path.exists(filename) and (overwrite is False):
            raise IOError("File %s already exists! To overwrite, set overwrite=True" % filename)

        with open(filename, "w") as yaml_file:
            yaml.dump(parameters, yaml_file,
                      indent=4,
                      explicit_start=True,
                      explicit_end=True)


# ## What is this??
# class ConfigurableYAML(type):
#
#     def __new__(cls, *args, **kwargs):
#
#         ConfigureYAML.constructors.append(cls)
#         return super(ConfigureableYAML, cls, *args, **kwargs)



import threading
import Queue
# from multiprocessing import Process, Queue
import datetime as dt
import logging
import numpy as np
from opyrant import hwio

logger = logging.getLogger(__name__)

class Events(object):
    """ Writes small event dictionaries out to a list of event handlers.

    Attributes
    ----------
    handlers: list
        The currently configured EventHandler instances

    Methods
    -------
    add_handler(handler) - appends the handler to the handlers list
    write(event) - Writes the dictionary `event` to each handler
    """

    def __init__(self):

        self.handlers = list()

    def add_handler(self, handler):
        """ Adds the handler to the list of handlers, as long as it supports
        writing.

        Parameters
        ----------
        handler: instance of EventHandler
            The handler to be added
        """

        if not hasattr(handler, "queue"):
            raise AttributeError("Event handler instance must contain a queue")

        self.handlers.append(handler)

    def close_handlers(self):
        """ Closes all of the existing handlers """

        for handler in self.handlers:
            handler.close()

    def write(self, event):
        """ Places the event in the queue for each handler to write.

        Parameters
        ----------
        event: dict
            A dictionary describing the current component event. It should have
            3 keys: name, action, and metadata. A time key will be added
            containing the datetime of the event.
        """
        if event is None:
            return

        event["time"] = dt.datetime.now()
        for handler in self.handlers:
            logger.debug("Adding to handler %s" % str(handler))
            handler.queue.put(event)


class EventHandler(object):
    """ Base class for all event handlers. Creates a separate thread that writes each event out using the handler's "write" method.

    Parameters
    ----------
    component: string
        Optionally argument that allows one to only log events with the
        specified component name.

    Attributes
    ----------
    thread: threading.Thread instance
        The thread that loops infinitely and writes each event placed in its
        queue out through the handler.
    queue: queue.Queue instance
        The queue that handles communicating events between threads.

    Methods
    -------
    write(event) - Writes the event using the specified handler
    close() - Ends the thread so everything can be properly closed out
    """
    STOP_FLAG = 0

    def __init__(self, component=None, *args, **kwargs):

        super(EventHandler, self).__init__(*args, **kwargs)

        self.component = component

        # Initialize the queue
        self.queue = Queue.Queue(maxsize=0)
        #self.queue = Queue(maxsize=0)

        # Initialize the thread
        self.thread = threading.Thread(target=self.run, name=self.__class__.__name__)
        # self.thread = Process(target=self.run, name=self.__class__.__name__)

        # Run the thread
        self.thread.start()

    def filter(self, event):
        """ Returns True if the event should be written """

        if self.component is None:
            return True

        return self.event["name"] == self.component

    def run(self):
        """ Runs inside the separate thread and calls the class' `write` method
        on any new events
        """
        while True:
            event = self.queue.get()
            if event is self.STOP_FLAG:
                logger.debug("Stopping thread %s" % self.thread.name)
                return
            if self.filter(event):
                self.write(event)

    def close(self):
        """ Ends the separate thread """

        self.queue.put(self.STOP_FLAG)

    def __del__(self):

        self.close()

    def write(self, event):

        raise NotImplementedError("Event handlers must implement a `write` method")


class EventInterfaceHandler(EventHandler, hwio.BooleanOutput):
    """ Handler to send event information out to a boolean interface. The event
    information is sent as a sequence of three chunks of bits, the first
    describing the name of the component, the second describing the action, and
    the third with any additional metadata. If the event details are too long to
    fit in the requested number of bytes, they are truncated first.

    Parameters
    ----------
    interface: instance of an Interface
        The interface to use to write the bit string out to hardware
    params: dictionary
        A set of key-value pairs that are sent to the interface when configuring
        the boolean write and when writing to it.
    name_bytes: int
        The number of bytes to use for encoding the name of the component
    action_bytes: int
        The number of bytes to use for encoding the action being performed
    metadata_bytes: int
        The number of bytes to use for any additional metadata.
    component: string
        Optionally argument that allows one to only log events with the
        specified component name.

    Attributes
    ----------
    thread: threading.Thread instance
        The thread that loops infinitely and writes each event placed in its
        queue out through the handler.
    queue: queue.Queue instance
        The queue that handles communicating events between threads.

    Methods
    -------
    write(event) - Writes the event using the specified handler
    close() - Ends the thread so everything can be properly closed out
    to_bit_sequence(event) - Serializes the event details into a string of bits
    """
    def __init__(self, interface, params={}, name_bytes=4, action_bytes=4,
                 metadata_bytes=16, component=None):

        self.name_bytes = name_bytes
        self.action_bytes = action_bytes
        self.metadata_bytes = metadata_bytes
        self.component = component
        self.map_to_bit = dict()
        super(EventInterfaceHandler, self).__init__(interface=interface,
                                                    params=params,
                                                    component=component)

    def write(self, event):
        """ Writes the event out the boolean output

        Parameters
        ----------
        event: dict
            A dictionary describing the current component event. It should have
            3 keys: name, action, and metadata.
        """

        try:
            key = (event["name"], event["action"], event["metadata"])
            bits = self.map_to_bit[key]
        except KeyError:
            bits = self.to_bit_sequence(event)
        self.interface._write_bool(value=bits, **self.params)

    def to_bit_sequence(self, event):
        """ Creates an array of bits containing the details in the event
        dictionary. Once created, the array is cached to speed up future writes.

        Parameters
        ----------
        event: dict
            A dictionary describing the current component event. It should have
            3 keys: name, action, and metadata.

        Returns
        -------
        The array of bits
        """

        if event["metadata"] is None:
            nbytes = self.action_bytes + self.name_bytes
            metadata_array = []
        else:
            nbytes = self.metadata_bytes  + self.action_bytes + self.name_bytes
            try:
                metadata_array = np.fromstring(event["metadata"],
                                               dtype=np.uint16).astype(np.uint8)[:self.metadata_bytes]
            except TypeError:
                metadata_array = np.array(map(ord,
                                              event["metadata"].ljust(self.metadata_bytes)[:self.metadata_bytes]),
                                          dtype=np.uint8)

        int8_array = np.zeros(nbytes, dtype="uint8")
        int8_array[:self.name_bytes] = map(ord, event["name"].ljust(self.name_bytes)[:self.name_bytes])
        int8_array[self.name_bytes:self.name_bytes + self.action_bytes] = map(ord, event["action"].ljust(self.action_bytes)[:self.action_bytes])
        int8_array[self.name_bytes + self.action_bytes:] = metadata_array

        sequence = ([True] +
                    np.unpackbits(int8_array).astype(bool).tolist() +
                    [False])
        key = (event["name"], event["action"], event["metadata"])
        self.map_to_bit[key] = sequence

        return sequence

    def toggle(self):
        pass


class EventDToAHandler(EventHandler):
    """ Handler to format event information so that it can be sent as a sequence
    of bits out an analog output. The event information is returned as a
    sequence of three chunks of bits, the first describing the name of the
    component, the second describing the action, and the third with any
    additional metadata. If the event details are too long to fit in the
    requested number of bytes, they are truncated first. Before being returned,
    the sequence is upsampled by a certain factor and then converted to float64.

    Parameters
    ----------
    name_bytes: int
        The number of bytes to use for encoding the name of the component
    action_bytes: int
        The number of bytes to use for encoding the action being performed
    metadata_bytes: int
        The number of bytes to use for any additional metadata.
    upsample_factor: int
        The factor by which the bit sequence should be upsampled.
    scaling: float
        A scaling factor to scale the analog representation of the digital signal (e.g. send out 3.3 Volts to pass to a digital input)
    component: string
        Optionally argument that allows one to only log events with the
        specified component name.

    All additional key-value pairs are stored for use by the interface

    Methods
    -------
    to_bit_sequence(event) - Serializes the event details into a string of bits
    """
    def __init__(self, name_bytes=4, action_bytes=4, metadata_bytes=16,
                 upsample_factor=1, scaling=1.0, component=None,
                 **interface_params):

        self.name_bytes = name_bytes
        self.action_bytes = action_bytes
        self.metadata_bytes = metadata_bytes
        self.upsample_factor = upsample_factor
        self.scaling = scaling
        self.component = component
        self.map_to_bit = dict()
        self.queue = Queue.Queue(maxsize=0)
        for key, value in interface_params.items():
            setattr(self, key, value)

    def filter(self, event):
        """ Always returns False, as this one should never be called by Events
        """

        return False

    def write(self, event):
        """ Does nothing """
        pass

    def to_bit_sequence(self, event):
        """ Creates an array of bits containing the details in the event
        dictionary. This array is then upsampled and converted to float64 to be
        sent down an analog output. Once created, the array is cached to speed
        up future calls.

        Parameters
        ----------
        event: dict
            A dictionary describing the current component event. It should have
            3 keys: name, action, and metadata.

        Returns
        -------
        The array of bits expressed as analog values
        """

        key = (event["name"], event["action"], event["metadata"])
        # Check if the bit string is already stored
        if key in self.map_to_bit:
            return self.map_to_bit[key]

        trim = lambda ss, l: ss.ljust(l)[:l]
        # Set up int8 arrays where strings are converted to integers using ord
        name_array = np.array(map(ord, trim(event["name"], self.name_bytes)),
                              dtype=np.uint8)
        action_array = np.array(map(ord, trim(event["action"],
                                              self.action_bytes)),
                                dtype=np.uint8)

        # Add the metadata array if a value was passed
        if event["metadata"] is not None:
            metadata_array = np.array(map(ord, trim(event["metadata"],
                                                    self.metadata_bytes)),
                                      dtype=np.uint8)
        else:
            metadata_array = np.array([], dtype=np.uint8)

        sequence = ([True] +
                    np.unpackbits(name_array).astype(bool).tolist() +
                    np.unpackbits(action_array).astype(bool).tolist() +
                    np.unpackbits(metadata_array).astype(bool).tolist() +
                    [False])
        sequence = np.repeat(sequence, self.upsample_factor).astype("float64")
        sequence *= self.scaling

        self.map_to_bit[key] = sequence

        return sequence

    def close(self):
        """ Nothing needs to be done """
        pass


class EventLogHandler(EventHandler):
    """ Writes event details out to a file log.

    Parameters
    ----------
    filename: string
        Path to the output file
    format: string
        A string that can be formatted with the event dictionary
    component: string
        Optional argument that allows one to only log events with the
        specified component name.

    Attributes
    ----------
    thread: threading.Thread instance
        The thread that loops infinitely and writes each event placed in its
        queue out through the handler.
    queue: queue.Queue instance
        The queue that handles communicating events between threads.

    Methods
    -------
    write(event) - Writes the event to the file
    close() - Ends the thread so everything can be properly closed out
    """
    def __init__(self, filename, format=None, component=None):

        self.filename = filename
        if format is None:
            self.format = "\t".join(["{time}",
                                     "{name}",
                                     "{action}",
                                     "{metadata}"])
        super(EventLogHandler, self).__init__(component=component)

    def write(self, event):
        """ Writes the event out to the file

        Parameters
        ----------
        event: dict
            A dictionary describing the current component event. It should have
            4 keys: name, action, and metadata added by the compnent, and time
            added by the Events class.
        """

        if "time" not in event:
            event["time"] = dt.datetime.now()

        with open(self.filename, "a") as fh:
            fh.write(self.format.format(**event) + "\n")

events = Events()

if __name__ == "__main__":

    ihandler = EventInterfaceHandler(None)
    events.add_handler(ihandler)
    for ii in range(100):
        events.write({})
        time.sleep(0.1)

    if ihandler.delay_queue.qsize() > 0:
        for ii in range(ihandler.delay_queue.qsize()):
            ihandler.delays.append(ihandler.delay_queue.get())

    print("Mean delay was %.4e seconds" % (sum(ihandler.delays) / 100))
    print("Max delay was %.4e seconds" % max(ihandler.delays))



import os
import csv
import logging
logger = logging.getLogger(__name__)


class Subject(object):
    """ Class which holds information about the subject currently running the
    experiment

    Parameters
    ----------
    name: string
        The name of the subject
    filename: string
        The name of the output file. The extension is used to determine the
        datastore type.

    Attributes
    ----------
    name: string
        The name of the subject
    filename: string
        The name of the output file
    datastore: Store object instance
        The datastore object in use

    Methods
    -------
    create_datastore(fields)
        Creates a datastore according to filename's extension
    store_data(trial)
        Stores a trial's data in the datastore
    """

    def __init__(self, name=None, filename=""):

        logger.debug("Creating subject object for %s" % name)
        self.name = name
        self.filename = filename
        logger.info("Created subject object with name %s" % self.name)
        self.datastore = None

    def create_datastore(self, fields):
        """ Creates a datastore object to store trial data

        Parameters
        ----------
        fields: list
            A list of field names to store from the trial object

        Returns
        -------
        bool
            True if the creation was successful

        Raises
        ------
        ValueError
            If filename extension is of unknown type
        """
        ext = os.path.splitext(self.filename)[1].lower()
        if ext == ".csv":
            self.datastore = CSVStore(fields, self.filename)
        else:
            raise ValueError("Extension %s is of unknown type" % ext)

        logger.info("Created datastore %s for subject %s" % (self.datastore,
                                                             self.name))

        return True

    def store_data(self, trial):
        """ Stores the trial data in the datastore

        Parameters
        ----------
        trial: instance of Trial
            The trial to store. It should have all fields used in creation of
            the datastore as attributes or annotations.

        Returns
        -------
        bool
            True if store succeeded
        """
        trial_dict = {}
        for field in self.datastore.fields:
            if hasattr(trial, field):
                trial_dict[field] = getattr(trial, field)
            elif field in trial.annotations:
                trial_dict[field] = trial.annotations[field]
            else:
                trial_dict[field] = None

        logger.debug("Storing data for trial %d" % trial.index)
        return self.datastore.store(trial_dict)


class CSVStore(object):
    """ Class that wraps storing trial data in a CSV file

    Parameters
    ----------
    fields: list
        A list of columns for the CSV file
    filename: string
        Full path to the csv file. Appends to the file if it already exists.

    Attributes
    ----------
    fields: list
        A list of columns for the CSV file
    filename: string
        Full path to the csv file

    Methods
    -------
    store(data)
        Appends data to the CSV file
    """
    def __init__(self, fields, filename):

        self.filename = filename
        self.fields = fields

        with open(self.filename, 'ab') as data_fh:
            trialWriter = csv.writer(data_fh)
            trialWriter.writerow(self.fields)

    def __str__(self):

        return "CSVStore: filename = %s, fields = %s" % (self.filename,
                                                         ", ".join(self.fields))

    def store(self, data):
        """ Appends the data to the CSV file

        Parameters
        ----------
        data: dictionary
            The data to store. The keys should match the fields specified when
            creating the CSVStore.

        Returns
        -------
        bool
            True if store succeeded
        """

        with open(self.filename, 'ab') as data_fh:
            trialWriter = csv.DictWriter(data_fh,
                                         fieldnames=self.fields,
                                         extrasaction='ignore')
            trialWriter.writerow(data)

        return True



import random
from opyrant.utils import rand_from_log_shape_dist
import cPickle as pickle
import numpy as np
import logging

logger = logging.getLogger(__name__)


def random_queue(items, weights=None, max_items=None):
    """ Generator which randomly samples items, with replacement

    Parameters
    ----------
    items: list
        A list of items to be queued
    weights: list
        A list of weights, 1 for each item in items
    max_items: int
        Maximum number of items to generate. (default: None)

    Yields
    ------
    A single item at each iteration
    """
    if len(items) == 0:
        raise ValueError("Cannot intialize a queue with 0 items")

    if weights is None:
        weights = [1.0 / len(items)] * len(items)
    else:
        weights = [float(ww) / sum(weights) for ww in weights]

    ii = 0
    while True:
        if (max_items is not None) and (ii >= max_items):
            break
        yield np.random.choice(items, p=weights)
        ii += 1


def block_queue(items, repetitions=1, shuffle=False):
    """ Generator which samples items in blocks

    Parameters
    ----------
    items: list
        A list of items to be queued
    repetitions: int
        The number of times each item in items will be presented (default: 1)
    shuffle: bool
        Shuffles the queue (default: False)

    Yields
    ------
    A single item at each iteration
    """
    items_repeated = []
    for rr in range(repetitions):
        items_repeated += items
    items = items_repeated

    if shuffle:
        random.shuffle(items)

    for item in items:
        yield item

class AdaptiveBase(object):
    """docstring for AdaptiveBase
    This is an abstract object for implementing adaptive procedures, such as
    a staircase. Importantly, any objects inheriting this need to define the
    `update()` and `next()` methods.
    """
    def __init__(self, **kwargs):
        self.updated = True # for first trial, no update needed
        self.update_error_str = "queue hasn't been updated since last trial"

    def __iter__(self):
        return self

    def update(self, correct, no_resp):
        self.updated = True
        if no_resp:
            self.no_response()

    def next(self):
        if not self.updated: #hasn't been updated since last trial
            raise Exception(self.update_error_msg()) #TODO: find what causes bug
        self.updated = False

    def no_response(self):
        pass

    def on_load(self):
        try:
            super(AdaptiveBase, self).on_load()
        except AttributeError:
            pass
        self.updated = True
        self.no_response()

    def update_error_msg(self):
        return self.update_error_str

class PersistentBase(object):
    """
    A mixin that allows for the creation of an obj through a load command that
    first checks for a pickled file to load an object before generating a new one.
    """
    def __init__(self, filename=None, **kwargs):
        assert filename != None
        super(PersistentBase, self).__init__(**kwargs)
        self.filename = filename
        self.save()

    @classmethod
    def load(cls, filename, *args, **kwargs):
        try:
            with open(filename, 'rb') as handle:
                ab = pickle.load(handle)
                ab.on_load()
            return ab
        except IOError:
            return cls(*args, filename=filename, **kwargs)

    def on_load(self):
        try:
            super(PersistentBase, self).on_load()
        except AttributeError:
            pass

    def save(self):
        with open(self.filename, 'wb') as handle:
            pickle.dump(self, handle)


class KaernbachStaircase(AdaptiveBase):
    """ generates values for a staircase procedure from Kaernbach 1991
    This procedure returns values for each trial and assumes that larger values are
    easier. Thus, after a correct trial, the next value returned will be smaller and
    after incorrect trials, the next value returned will be larger. The magnitudes of
    these changes are stepsize_dn and stepsize_up, respectively.
    Args:
        start_val (float/int): the starting value of the procedure (default: 100)
    Kwargs:
        stepsize_up (int): number of steps to take after incorrect trial (default: 3)
        stepsize_dn (int): number of steps to take after correct trial (default: 1)
        min_val (float): minimum parameter value to allow (default: 0)
        max_val (float): maximum parameter value to allow (default: 100)
        crit (int): minimum number of trials (default: 0)
        crit_method (int): maximum number of trials (default: 100)
    Returns:
        float
    """
    def __init__(self,
                 start_val=100,
                 stepsize_up=3,
                 stepsize_dn=1,
                 min_val=0,
                 max_val=100,
                 crit=100,
                 crit_method='trials'
                 ):
        super(KaernbachStaircase, self).__init__()
        self.val = start_val
        self.stepsize_up = stepsize_up
        self.stepsize_dn = stepsize_dn
        self.min_val = min_val
        self.max_val = max_val
        self.crit = crit
        self.crit_method = crit_method
        self.counter = 0
        self.going_up = False

    def update(self, correct, no_resp):
        super(KaernbachStaircase, self).update(correct, no_resp)

        self.val += -1*self.stepsize_dn if correct else self.stepsize_up

        if self.crit_method=='reversals':
            if correct==self.going_up: # checks if last trial's perf was consistent w/ trend
                self.counter += 1
                self.going_up = not self.going_up

        # stop at max/min if we hit the rails
        if (self.max_val!=None) and (self.val > self.max_val):
            self.val = self.max_val
        elif (self.min_val!=None) and (self.val < self.min_val):
            self.val = self.min_val

    def next(self):
        super(KaernbachStaircase, self).next()
        if self.counter > self.crit:
            raise StopIteration
        self.counter += 1 if self.crit_method=='trials' else 0
        return self.val

class DoubleStaircase(AdaptiveBase):
    """
    Generates conditions from a list of stims that monotonically vary from most
    easily left to most easily right
    i.e. left is low and right is high

    The goal of this queue is to estimate the 50% point of a psychometric curve.

    This will probe left and right trials, if the response is correct, it will
    move the indices closer to each other until they are adjacent.

    stims: an array of stimuli names ordered from most easily left to most easily right
    rate_constant: the step size is the rate_constant*(high_idx-low_idx)
    """
    def __init__(self, stims, rate_constant=.05, **kwargs):
        super(DoubleStaircase, self).__init__(**kwargs)
        self.stims = stims
        self.rate_constant = rate_constant
        self.low_idx = 0
        self.high_idx = len(self.stims) - 1
        self.trial = {}
        self.update_error_str = "double staircase queue %s hasn't been updated since last trial" % (self.stims[0])

    def update(self, correct, no_resp):
        super(DoubleStaircase, self).update(correct, no_resp)
        if correct:
            if self.trial['low']:
                self.low_idx = self.trial['value']
            else:
                self.high_idx = self.trial['value']
        self.trial = {}

    def next(self):
        super(DoubleStaircase, self).next()
        if self.high_idx - self.low_idx <= 1:
            raise StopIteration

        delta = int(np.ceil((self.high_idx - self.low_idx) * self.rate_constant))
        if random.random() < .5: # probe low side
            self.trial['low'] = True
            self.trial['value'] = self.low_idx + delta
            return {'class': 'L',  'stim_name': self.stims[self.trial['value']]}
        else:
            self.trial['low'] = False
            self.trial['value'] = self.high_idx - delta
            return {'class': 'R',  'stim_name': self.stims[self.trial['value']]}

    def no_response(self):
        super(DoubleStaircase, self).no_response()
        self.trial = {}

    def update_error_msg(self):
        sup = super(DoubleStaircase, self).update_error_msg()
        state = "self.trial.low=%s    self.trial.value=%d    self.low_idx=%d    self.high_idx=%d" % (self.trial['low'], self.trial['value'], self.low_idx, self.high_idx)
        return "\n".join([sup, state])

class DoubleStaircaseReinforced(AdaptiveBase):
    """
    Generates conditions as with DoubleStaircase, but 1-probe_rate proportion of
    the trials easier/known trials to reduce frustration.

    Easier trials are sampled from a log shaped distribution so that more trials
    are sampled from the edges than near the indices

    stims: an array of stimuli names ordered from most easily left to most easily right
    rate_constant: the step size is the rate_constant*(high_idx-low_idx)
    probe_rate: proportion of trials that are between [0, low_idx] or [high_idx, length(stims)]
    """
    def __init__(self, stims, rate_constant=.05, probe_rate=.1, sample_log=False, **kwargs):
        super(DoubleStaircaseReinforced, self).__init__(**kwargs)
        self.dblstaircase = DoubleStaircase(stims, rate_constant)
        self.stims = stims
        self.probe_rate = probe_rate
        self.sample_log = sample_log
        self.last_probe = False
        self.update_error_str = "reinforced double staircase queue %s hasn't been updated since last trial" % (self.stims[0])

    def update(self, correct, no_resp):
        if self.last_probe:
            self.dblstaircase.update(correct, no_resp)
        super(DoubleStaircaseReinforced, self).update(correct, no_resp)

    def next(self):
        super(DoubleStaircaseReinforced, self).next()

        if random.random() < self.probe_rate:
            try:
                ret = self.dblstaircase.next()
                self.last_probe = True
                return ret
            except StopIteration:
                self.probe_rate = 0
                self.last_probe = False
                self.updated = True
                return self.next()
        else:
            self.last_probe = False
            if random.random() < .5: # probe left
                if self.sample_log:
                    val = int((1 - rand_from_log_shape_dist()) * self.dblstaircase.low_idx)
                else:
                    val = random.randrange(self.dblstaircase.low_idx + 1)
                return {'class': 'L',  'stim_name': self.stims[val]}
            else: # probe right
                if self.sample_log:
                    val = self.dblstaircase.high_idx + int(rand_from_log_shape_dist() * (len(self.stims) - self.dblstaircase.high_idx))
                else:
                    val = self.dblstaircase.high_idx + random.randrange(len(self.stims) - self.dblstaircase.high_idx)
                return {'class': 'R',  'stim_name': self.stims[val]}

    def no_response(self):
        super(DoubleStaircaseReinforced, self).no_response()

    def on_load(self):
        super(DoubleStaircaseReinforced, self).on_load()
        self.dblstaircase.on_load()

    def update_error_msg(self):
        sup = super(DoubleStaircaseReinforced, self).update_error_msg()
        state = "self.last_probe=%s" % (self.last_probe)
        sub_state = "self.dbs.trial=%s    self.dbs.low_idx=%d    self.dbs.high_idx=%d" % (self.dblstaircase.trial, self.dblstaircase.low_idx, self.dblstaircase.high_idx)
        return "\n".join([sup, state, sub_state])


class MixedAdaptiveQueue(PersistentBase, AdaptiveBase):
    """
    Generates conditions from multiple adaptive sub queues.

    Use the generator MixedAdaptiveQueue.load(filename, sub_queues)
    to load a previously saved MixedAdaptiveQueue or generate a new one
    if the pkl file doesn't exist.

    sub_queues: a list of adaptive queues
    probabilities: a list of weights with which to sample from sub_queues
                        should be same length as sub_queues
                        NotImplemented
    filename: filename of pickle to save itself
    """
    def __init__(self, sub_queues, probabilities=None, **kwargs):
        super(MixedAdaptiveQueue, self).__init__(**kwargs)
        self.sub_queues = sub_queues
        self.probabilities = probabilities
        self.sub_queue_idx = -1
        self.update_error_str = "MixedAdaptiveQueue hasn't been updated since last trial"
        self.save()

    def update(self, correct, no_resp):
        super(MixedAdaptiveQueue, self).update(correct, no_resp)
        self.sub_queues[self.sub_queue_idx].update(correct, no_resp)
        self.save()

    def next(self):
        super(MixedAdaptiveQueue, self).next()
        if self.probabilities is None:
            try:
                self.sub_queue_idx = random.randrange(len(self.sub_queues))
                return self.sub_queues[self.sub_queue_idx].next()
            except StopIteration:
                #TODO: deal with subqueue finished, and possibility of all subqueues finishing
                raise NotImplementedError
        else:
            #TODO: support variable probabilities for each sub_queue
            raise NotImplementedError

    def on_load(self):
        super(MixedAdaptiveQueue, self).on_load()
        for sub_queue in self.sub_queues:
            try:
                sub_queue.on_load()
            except AttributeError:
                pass


class BaseHandler(object):
    """ Base class for implementing an iterable queue handler

    Parameters
    ----------
    queue: queue function or class
        The queue that will be iterated over. All queues must accept an items
        argument and implement generator, either through yielding values or a
        Class.next() method.
    items: list
        A list of items to iterate over.
    Additional key-value pairs are used to initialize the queue

    Attributes
    ----------
    queue: queue generator or class instance
        The queue that will be iterated over.
    queue_parameters: dict
        All additional parameters used to initialize the queue.        
    """

    def __init__(self, queue, items, **queue_parameters):

        if not hasattr(queue, "__call__"):
            raise TypeError("queue must be a callable function or class")
        # Store these in case we need to reset
        self._queue = queue
        self._items = items

        self.queue = queue(items=items, **queue_parameters)
        self.queue_parameters = queue_parameters

    def reset(self):
        """ Reset the queue """

        self.queue = self._queue(items=self._items, **self.queue_parameters)

    def __iter__(self):

        for item in self.queue:
            yield item



import logging
import datetime as dt
from opyrant import EndSession
from opyrant.events import events

logger = logging.getLogger(__name__)


class Trial(object):
    """ Class that implements all basic functionality of a trial

    Parameters
    ----------
    index: int
        Index of the trial
    experiment: instance of Experiment class
        The experiment of which this trial is a part
    block: instance of Block class
        The block that generated this trial
    condition: instance of StimulusCondition
        The condition for the current trial. Provides the trial with a stimulus,
        as well as reinforcement instructions

    Attributes
    ----------
    index: int
        Index of the trial
    experiment: instance of Experiment class
        The experiment of which this trial is a part
    stimulus_condition: instance of StimulusCondition
        The condition for the current trial. Provides the trial with a stimulus,
        as well as reinforcement instructions
    time: datetime
        The time the trial started
    session: int
        Index of the current session

    Methods
    -------
    run() - Runs the trial
    annotate() - Annotates the trial with key-value pairs
    """
    def __init__(self,
                 index=None,
                 experiment=None,
                 condition=None,
                 block=None,
                 *args, **kwargs):

        super(Trial, self).__init__(*args, **kwargs)

        # Object references
        self.experiment = experiment
        self.condition = condition
        self.block = block
        self.annotations = dict()

        # Trial properties
        self.index = index
        self.session = self.experiment.session_id
        self.time = None  # Set just after trial_pre

        # Likely trial details
        self.stimulus = None
        self.response = None
        self.rt = None
        self.correct = False
        self.reward = False
        self.punish = False

        # Trial event information
        self.event = dict(name="Trial",
                          action="",
                          metadata="")

    def annotate(self, **annotations):
        """ Annotate the trial with key-value pairs """

        self.annotations.update(annotations)

    def run(self):
        """ Runs the trial

        Summary
        -------
        The main structure is as follows:

        Get stimulus -> Initiate trial -> Play stimulus -> Receive response ->
        Consequate response -> Finish trial -> Save data.

        The stimulus, response and consequate stages are broken into pre, main,
        and post stages. Only use the stages you need in your experiment.
        """

        self.experiment.this_trial = self

        # Get the stimulus
        self.stimulus = self.condition.get()

        # Any pre-trial logging / computations
        self.experiment.trial_pre()

        # Emit trial event
        self.event.update(action="start", metadata=str(self.index))
        events.write(self.event)

        # Record the trial time
        self.time = dt.datetime.now()

        # Perform stimulus playback
        self.experiment.stimulus_pre()
        self.experiment.stimulus_main()
        self.experiment.stimulus_post()

        # Evaluate subject's response
        self.experiment.response_pre()
        self.experiment.response_main()
        self.experiment.response_post()

        # Consequate the response with a reward, punishment or neither
        if self.response == self.condition.response:
            self.correct = True
            if self.condition.is_rewarded and self.block.reinforcement.consequate(self):
                self.reward = True
                self.experiment.reward_pre()
                self.experiment.reward_main()
                self.experiment.reward_post()
        else:
            self.correct = False
            if self.condition.is_punished and self.block.reinforcement.consequate(self):
                self.punish = True
                self.experiment.punish_pre()
                self.experiment.punish_main()
                self.experiment.punish_post()

        # Emit trial end event
        self.event.update(action="end", metadata=str(self.index))
        events.write(self.event)

        # Finalize trial
        self.experiment.trial_post()

        # Store trial data
        self.experiment.subject.store_data(self)

        # Update session schedulers
        self.experiment.session.update()

        if self.experiment.check_session_schedule() is False:
            logger.debug("Session has run long enough. Ending")
            raise EndSession



import logging
import datetime as dt
import numpy as np
from opyrant import (EndSession,
                     EndExperiment,
                     ComponentError,
                     InterfaceError,
                     utils)

logger = logging.getLogger(__name__)


class State(object):
    """ States provide a nice interface for running experiments and transitioning to sleep/idle phases. By implementing __enter__ and __exit__ methods, they use the "with" statement construct that allows for simple error handling (e.g. session ends, keyboard interrupts to stop an experiment, etc.)

    Parameters
    ----------
    schedulers: list of scheduler objects
        These determine whether or not the state should be running, using their check method

    Methods
    -------
    check() - Check if the state should be active according to its schedulers
    run() - Run the state (should be used within the "with" statement)
    start() - wrapper for run that includes the "with" statement
    """

    def __init__(self, experiment=None, schedulers=None):

        if schedulers is None:
            schedulers = list()
        if not isinstance(schedulers, list):
            schedulers = [schedulers]
        self.schedulers = schedulers
        self.experiment = experiment

    def check(self):
        """ Checks all of the states schedulers to see if the state should be active.

        Returns
        -------
        True if the state should be active and False otherwise.
        """

        # If any scheduler says not to run, then don't run
        for scheduler in self.schedulers:
            if not scheduler.check():
                return False

        return True

    def __enter__(self):
        """ Start all of the schedulers """

        logger.info("Entering %s state" % self.__class__.__name__)
        for scheduler in self.schedulers:
            scheduler.start()

        return self

    def __exit__(self, type_, value, traceback):
        """ Handles KeyboardInterrupt and EndExperiment exceptions to end the experiment, EndSession exceptions to end the session state, and logs all others.
        """
        logger.info("Exiting %s state" % self.__class__.__name__)

        # Stop the schedulers
        for scheduler in self.schedulers:
            scheduler.stop()

        # Handle expected exceptions
        if type_ in [KeyboardInterrupt, EndExperiment]:
            logger.info("Finishing experiment")
            self.experiment.end()
            return True
        elif type_ is EndSession:
            logger.info("Session has ended")
            return True

        # Log all other exceptions and raise them
        if isinstance(value, Exception):
            if type_ in [InterfaceError, ComponentError]:
                logger.critical("There was a critical error in communicating with the hardware!")
            logger.critical(repr(value))

        return False

    def run(self):

        pass

    def start(self):
        """ Implements the "with" context for this state """

        with self as state:
            state.run()


class Session(State):
    """ Session state for running an experiment. Should be used with the "with" statement (see Examples).

    Parameters
    ----------
    schedulers: list of scheduler objects
        These determine whether or not the state should be running, using their check method
    experiment: an instance of a Behavior class
        The experiment whose session methods should be run.

    Methods
    -------
    check() - Check if the state should be active according to its schedulers
    run() - Run the experiment's session_main method
    start() - wrapper for run that includes the "with" statement
    update() - Update schedulers at the end of the trial

    Examples
    --------
    with Session(experiment=experiment) as state: # Runs experiment.session_pre
        state.run() # Runs experiment.session_main
    # Exiting with statement runs experiment.session_post

    # "with" context is also implemented in the start() method
    state = Session(experiment=experiment)
    state.start()
    """

    def __enter__(self):

        self.experiment.session_pre()
        for scheduler in self.schedulers:
            scheduler.start()

        return self

    def run(self):
        """ Runs session main """

        self.experiment.session_main()

    def __exit__(self, type_, value, traceback):

        self.experiment.session_post()

        return super(Session, self).__exit__(type_, value, traceback)

    def update(self):
        """ Updates all schedulers with information on the current trial """

        if hasattr(self.experiment, "this_trial"):
            for scheduler in self.schedulers:
                scheduler.update(self.experiment.this_trial)


class Idle(State):
    """ A simple idle state.

    Parameters
    ----------
    experiment: an instance of a Behavior class
        The experiment whose session methods should be run.
    poll_interval: int
        The interval, in seconds, at which other states should be checked to run

    Methods
    -------
    run() - Run the experiment's session_main method
    """
    def __init__(self, experiment=None, poll_interval=60):

        super(Idle, self).__init__(experiment=experiment,
                                   schedulers=None)
        self.poll_interval = poll_interval

    def run(self):
        """ Checks if the experiment should be sleeping or running a session and kicks off those states. """

        while True:
            if self.experiment.check_sleep_schedule():
                return self.experiment._sleep.start()
            elif self.experiment.check_session_schedule():
                return self.experiment.session.start()
            else:
                logger.debug("idling...")
                utils.wait(self.poll_interval)


class Sleep(State):
    """ A panel sleep state. Turns off all outputs, checking every so often if it should wake up

    Parameters
    ----------
    experiment: an instance of a Behavior class
        The experiment whose session methods should be run.
    schedulers: an instance of TimeOfDayScheduler
        The time of day scheduler to follow for when to sleep.
    poll_interval: int
        The interval, in seconds, at which other states should be checked to run
    time_period: string or tuple
        Either "night" or a tuple of "HH:MM" start and end times. Only used if scheduler is not provided.

    Methods
    -------
    run() - Run the experiment's session_main method
    """
    def __init__(self, experiment=None, schedulers=None, poll_interval=60,
                 time_period="night"):

        if schedulers is None:
            schedulers = TimeOfDayScheduler(time_period)
        self.poll_interval = poll_interval

        super(Sleep, self).__init__(experiment=experiment,
                                    schedulers=schedulers)

    def run(self):
        """ Checks every poll interval whether the panel should be sleeping and puts it to sleep """

        while True:
            logger.debug("sleeping")
            self.experiment.panel.sleep()
            utils.wait(self.poll_interval)
            if not self.check():
                break
        self.experiment.panel.wake()


class BaseScheduler(object):
    """ Implements a base class for scheduling states

    Summary
    -------
    Schedulers allow the state to be started and stopped based on certain critera. For instance, you can start the sleep state when the sun sets, or stop and session state after 100 trials.

    Methods
    -------
    check() - Checks whether the state should be active
    start() - Run when the state starts to initialize any variables
    stop() - Run when the state finishes to close out any variables
    update(trial) - Run after each trial to update the scheduler if necessary
    """

    def __init__(self):

        pass

    def start(self):

        pass

    def stop(self):

        pass

    def update(self, trial):

        pass

    def check(self):
        """ This should really be implemented by the subclass """

        raise NotImplementedError("Scheduler %s does not have a check method" % self.__class__.__name__)


class TimeOfDayScheduler(BaseScheduler):
    """ Schedule a state to start and stop depending on the time of day

    Parameters
    ----------
    time_periods: string or list
        The time periods in which this schedule should be active. The value of "sun" can be passed to use the current day-night schedule. Otherwise, pass a list of tuples (start, end) (e.g. [("5:00", "17:00")] for 5am to 5pm)

    Methods
    -------
    check() - Returns True if the state should be active according to this schedule
    """

    def __init__(self, time_periods="sun"):

        # Any other sanitizations?
        if isinstance(time_periods, tuple):
            time_periods = [time_periods]
        self.time_periods = time_periods

    def check(self):
        """ Returns True if the state should be active according to this schedule
        """

        return utils.check_time(self.time_periods)


class TimeScheduler(BaseScheduler):
    """ Schedules a state to start and stop based on how long the state has been active and how long since the state was previously active.

    Parameters
    ----------
    duration: int
        The duration, in minutes, that the state should be active
    interval: int
        The time since the state was last active before it should become active again.

    Methods
    -------
    start() - Stores the start time of the current state
    stop() - Stores the end time of the current state
    check() - Returns True if the state should activate
    """
    def __init__(self, duration=None, interval=None):

        self.duration = duration
        self.interval = interval

        self.start_time = None
        self.stop_time = None

    def start(self):
        """ Stores the start time of the current state """

        self.start_time = dt.datetime.now()
        self.stop_time = None

    def stop(self):
        """ Stores the end time of the current state """

        self.stop_time = dt.datetime.now()
        self.start_time = None

    def check(self):
        """ Checks if the current time is greater than `duration` minutes after start time or `interval` minutes after stop time """

        current_time = dt.datetime.now()
        # If start_time is None, the state is not active. Should it be?
        if self.start_time is None:
            # No interval specified, always start
            if self.interval is None:
                return True

            # The state hasn't activated yet, always start
            if self.stop_time is None:
                return True

            # Has it been greater than interval minutes since the last time?
            time_since = (current_time - self.stop_time).total_seconds() / 60.
            if time_since < self.interval:
                return False

        # If stop_time is None, the state is currently active. Should it stop?
        if self.stop_time is None:
            # No duration specified, so do not stop
            if self.duration is None:
                return True

            # Has the state been active for long enough?
            time_since = (current_time - self.start_time).total_seconds() / 60.
            if time_since >= self.duration:
                return False

        return True


class CountScheduler(BaseScheduler):
    """ Schedules a state stop after a certain number of trials.

    Parameters
    ----------
    max_trials: int
        The maximum number of trials

    Methods
    -------
    check() - Returns True if the state has not yet reached max_trials

    TODO: This could be expanded to include things like total number of rewards or correct responses.
    """
    def __init__(self, max_trials=None):

        self.max_trials = max_trials
        self.trial_index = 0

    def check(self):
        """ Returns True if current trial index is less than max_trials """

        if self.max_trials is None:
            return True

        return self.trial_index < self.max_trials

    def stop(self):
        """ Resets the trial index since the session is over """

        self.trial_index = 0

    def update(self, trial):
        """ Updates the current trial index """

        self.trial_index = trial.index


available_states = {"idle": Idle,
                    "session": Session,
                    "sleep": Sleep}

available_schedulers = {"day": TimeOfDayScheduler,
                        "timeofday": TimeOfDayScheduler,
                        "time": TimeScheduler}
